from __future__ import division, print_function, absolute_import
import unittest
import numpy as num
from numpy.testing import assert_allclose

from pyrocko.dataset import tectonics
from pyrocko import util


class TectonicsTestCase(unittest.TestCase):

    def test_boundaries(self):
        bird = tectonics.PeterBird2003()
        boundaries = bird.get_boundaries()
        assert len(boundaries) == 229

    def test_boundaries_types(self):
        bird = tectonics.PeterBird2003()
        boundaries = bird.get_boundaries()
        for boundary in boundaries:
            if (boundary.name1, boundary.kind, boundary.name2) == (
                    'PS', '/', 'PA'):
                tt = boundary.split_types()
                assert len(tt) == 1
                assert tt[0][0] == 'SUB'

            if (boundary.name1, boundary.kind, boundary.name2) == (
                    'AS', '-', 'AT'):

                lastpoint = None
                for typ, part in boundary.split_types():
                    if lastpoint is not None:
                        assert_allclose(lastpoint, part[0, :])
                        lastpoint = part[:, -1]

                lastpoint = None
                types = []
                for typ, part in boundary.split_types(
                        [['OSR'],
                         ['SUB'],
                         ['OTF', 'OCB', 'CTF', 'CCB', 'CRB']]):

                    if lastpoint is not None:
                        assert_allclose(lastpoint, part[0, :])
                        lastpoint = part[:, -1]

                    types.append(typ)

                assert types == [['OSR'], ['CTF', 'CRB']]
            boundary.split_types()

    def test_plates(self):
        bird = tectonics.PeterBird2003()
        plates = bird.get_plates()

        assert len(plates) == 52

        point = num.array([-25., 135.], dtype=num.float)
        for plate in plates:
            plate.max_interpoint_distance()
            assert (plate.name == 'AU') == plate.contains_point(point)

        lats = num.linspace(-20., -10., 20)
        lons = num.linspace(-80., -70., 20)
        lats2 = num.repeat(lats, lons.size)
        lons2 = num.tile(lons, lats.size)
        points = num.vstack((lats2, lons2)).T

        full_names = []
        for plate in plates:
            if num.any(plate.contains_points(points)):
                full_names.append(bird.full_name(plate.name))

        assert full_names == [
            'South American Plate',
            'Nazca Plate',
            'Altiplano Plate']

    def test_velocities(self):
        gsrm = tectonics.GSRM1()
        lats, lons, vnorth, veast, vnorth_err, veast_err, corr = \
            gsrm.get_velocities('AF', region=(20, 30, 0., 10.))
        assert 0.0 < num.mean(num.sqrt(veast**2 + vnorth**2)) < 0.01
        assert_allclose(
            [lats.min(), lats.max(), lons.min(), lons.max()],
            [0., 10., 20., 30.], atol=1e-5)

        assert gsrm.full_name('NU') == 'African Plate'
        assert 'EU' in gsrm.plate_names()


if __name__ == "__main__":
    util.setup_logging('test_tectonics', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import math
from time import time
import unittest
import logging
from tempfile import mkdtemp
import numpy as num
import os
import shutil

from pyrocko import orthodrome as ortd
from pyrocko import util, gf, cake  # noqa
from pyrocko.fomosto import psgrn_pscmp
from .common import Benchmark

logger = logging.getLogger('pyrocko.test.test_gf_psgrn_pscmp')
benchmark = Benchmark()

km = 1000.

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.


def statics(engine, source, starget):
    store = engine.get_store(starget.store_id)
    dsource = source.discretize_basesource(store, starget)

    assert len(starget.north_shifts) == len(starget.east_shifts)

    out = num.zeros((len(starget.north_shifts), len(starget.components)))
    sfactor = source.get_factor()
    for i, (north, east) in enumerate(
            zip(starget.north_shifts, starget.east_shifts)):

        receiver = gf.Receiver(
            lat=starget.lat,
            lon=starget.lon,
            north_shift=north,
            east_shift=east,
            depth=starget.depth)

        values = store.statics(
            dsource, receiver, starget.components,
            interpolation=starget.interpolation)

        for icomponent, value in enumerate(values):
            out[i, icomponent] = value * sfactor

    return out


@unittest.skipUnless(
    psgrn_pscmp.have_backend(), 'backend psgrn_pscmp not available')
class GFPsgrnPscmpTestCase(unittest.TestCase):

    tempdirs = []

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def test_fomosto_vs_psgrn_pscmp(self):

        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
mantle
 35. 8.04 4.48 3.58 1449. 600.
 77.5 8.045 4.49 3.5 1445. 600.
 77.5 8.045 4.49 3.5 180.6 75.
 120. 8.05 4.5 3.427 180. 75.
 120. 8.05 4.5 3.427 182.6 76.06
 165. 8.175 4.509 3.371 188.7 76.55
 210. 8.301 4.518 3.324 201. 79.4
 210. 8.3 4.52 3.321 336.9 133.3
 410. 9.03 4.871 3.504 376.5 146.1
 410. 9.36 5.08 3.929 414.1 162.7
 660. 10.2 5.611 3.918 428.5 172.9
 660. 10.79 5.965 4.229 1349. 549.6'''.lstrip()))

        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)
        store_id = 'psgrn_pscmp_test'

        version = '2008a'

        c = psgrn_pscmp.PsGrnPsCmpConfig()
        c.psgrn_config.sampling_interval = 1.
        c.psgrn_config.version = version
        c.pscmp_config.version = version

        config = gf.meta.ConfigTypeA(
            id=store_id,
            ncomponents=10,
            sample_rate=1. / (3600. * 24.),
            receiver_depth=0. * km,
            source_depth_min=0. * km,
            source_depth_max=5. * km,
            source_depth_delta=0.1 * km,
            distance_min=0. * km,
            distance_max=40. * km,
            distance_delta=0.1 * km,
            modelling_code_id='psgrn_pscmp.%s' % version,
            earthmodel_1d=mod,
            tabulated_phases=[])

        config.validate()
        gf.store.Store.create_editables(
            store_dir, config=config, extra={'psgrn_pscmp': c})

        store = gf.store.Store(store_dir, 'r')
        store.close()

        # build store
        try:
            psgrn_pscmp.build(store_dir, nworkers=1)
        except psgrn_pscmp.PsCmpError as e:
            if str(e).find('could not start psgrn/pscmp') != -1:
                logger.warn('psgrn/pscmp not installed; '
                            'skipping test_pyrocko_gf_vs_pscmp')
                return
            else:
                raise

        origin = gf.Location(
            lat=10.,
            lon=-15.)

        # test GF store
        TestRF = dict(
            lat=origin.lat,
            lon=origin.lon,
            depth=2. * km,
            width=2. * km,
            length=5. * km,
            rake=90., dip=45., strike=45.,
            slip=1.)

        source = gf.RectangularSource(**TestRF)

        neast = 40
        nnorth = 40

        N, E = num.meshgrid(num.linspace(-20. * km, 20. * km, nnorth),
                            num.linspace(-20. * km, 20. * km, neast))

        starget = gf.StaticTarget(
            lats=num.array([origin.lat] * N.size),
            lons=num.array([origin.lon] * N.size),
            north_shifts=N.flatten(),
            east_shifts=E.flatten(),
            interpolation='nearest_neighbor')
        engine = gf.LocalEngine(store_dirs=[store_dir])
        t0 = time()
        r = engine.process(source, starget)
        t1 = time()
        logger.info('pyrocko stacking time %f' % (t1 - t0))
        un_fomosto = r.static_results()[0].result['displacement.n']
        ue_fomosto = r.static_results()[0].result['displacement.e']
        ud_fomosto = r.static_results()[0].result['displacement.d']

        # test against direct pscmp output
        lats, lons = ortd.ne_to_latlon(
            origin.lat, origin.lon, N.flatten(), E.flatten())
        pscmp_sources = [psgrn_pscmp.PsCmpRectangularSource(**TestRF)]

        cc = c.pscmp_config
        cc.observation = psgrn_pscmp.PsCmpScatter(lats=lats, lons=lons)

        cc.rectangular_source_patches = pscmp_sources

        ccf = psgrn_pscmp.PsCmpConfigFull(**cc.items())
        ccf.psgrn_outdir = os.path.join(store_dir, c.gf_outdir) + '/'

        t2 = time()
        runner = psgrn_pscmp.PsCmpRunner(keep_tmp=False)
        runner.run(ccf)
        ps2du = runner.get_results(component='displ')[0]
        t3 = time()
        logger.info('pscmp stacking time %f' % (t3 - t2))

        un_pscmp = ps2du[:, 0]
        ue_pscmp = ps2du[:, 1]
        ud_pscmp = ps2du[:, 2]

        num.testing.assert_allclose(un_fomosto, un_pscmp, atol=0.002)
        num.testing.assert_allclose(ue_fomosto, ue_pscmp, atol=0.002)
        num.testing.assert_allclose(ud_fomosto, ud_pscmp, atol=0.002)

        # plotting

#        uz_min = num.min(uz)
#        uz_max = num.max(uz)
#        uz2d_min = num.min(uz2d)
#        uz2d_max = num.max(uz2d)

#        uz_absmax = max(abs(uz_min), abs(uz_max))
#        uz2d_absmax = max(abs(uz2d_min), abs(uz2d_max))

#        levels = num.linspace(-uz_absmax, uz_absmax, 21)
#        levels2d = num.linspace(-uz2d_absmax, uz2d_absmax, 21)

#        from matplotlib import pyplot as plt

#        fontsize = 10.
#        plot.mpl_init(fontsize=fontsize)

#        cmap = plt.cm.get_cmap('coolwarm')
#        fig = plt.figure(figsize=plot.mpl_papersize('a4', 'landscape'))
#        plot.mpl_margins(fig, w=14., h=6., units=fontsize)

#        axes1 = fig.add_subplot(1, 2, 1, aspect=1.0)
#        cs1 = axes1.contourf(
#            easts / km, norths / km, uz, levels=levels, cmap=cmap)
#        plt.colorbar(cs1)

#        axes2 = fig.add_subplot(1, 2, 2, aspect=1.0)
#        cs2 = axes2.contourf(
#            easts / km, norths / km, uz2d, levels=levels2d, cmap=cmap)
#        plt.colorbar(cs2)

#        axes1.set_xlabel('Easting [km]')
#        axes1.set_ylabel('Northing [km]')

#        fig.savefig('staticGFvs2d_Afmu_diff.pdf')


if __name__ == '__main__':
    util.setup_logging('test_gf_psgrn_pscmp', 'warning')
    unittest.main()



def _run_main(app, arguments_list):
    for args in arguments_list:
        try:
            app.main(args=args)
        except SystemExit:
            pass


def test_fomosto():
    from pyrocko.apps import fomosto
    import tempfile
    import shutil

    tmp_store = tempfile.mkdtemp(prefix='pyrocko')

    test_arguments = (
        ['--help'],
        ['init', 'qseis.2006a', tmp_store]
        )

    _run_main(fomosto, test_arguments)

    shutil.rmtree(tmp_store)


def test_cake():
    from pyrocko.apps import cake

    test_arguments = (
        ['--help'],
        ['print'],
        ['arrivals'],
        ['paths'],
        ['plot-xt'],
        ['plot-xp'],
        ['plot-rays'],
        ['plot'],
        ['plot-model'],
        ['list-models'],
        ['list-phase-map'],
        ['simplify-model'],
        ['scatter'],
        )

    _run_main(cake, test_arguments)


def test_hamster():
    from pyrocko.apps import hamster

    test_arguments = (
        ['--help'],
        )

    _run_main(hamster, test_arguments)


def test_jackseis():
    from pyrocko.apps import jackseis

    test_arguments = (
        ['--help'],
        )

    _run_main(jackseis, test_arguments)


def test_snuffler():
    from pyrocko.apps import snuffler

    test_arguments = (
        ['--help'],
        )

    _run_main(snuffler, test_arguments)



from __future__ import division, print_function, absolute_import
from builtins import range

import unittest
import numpy as num
from pyrocko import gf, util

km = 1e3


class GFSourceTypesTestCase(unittest.TestCase):

    def test_rectangular_source(self):
        # WIP
        nsrc = 5
        rect_sources = []

        for n in range(nsrc):
            src = gf.RectangularSource(
                lat=0., lon=0.,
                anchor='bottom',
                north_shift=5000., east_shift=9000., depth=4.*km,
                width=2.*km, length=8.*km,
                dip=0., rake=0., strike=(180./nsrc + 1) * n,
                slip=1.)
            rect_sources.append(src)

    @staticmethod
    def plot_rectangular_source(src, store):
        from matplotlib import pyplot as plt
        from matplotlib.patches import Polygon
        ax = plt.gca()
        ne = src.outline(cs='xy')
        p = Polygon(num.fliplr(ne), fill=False, color='r', alpha=.7)
        ax.add_artist(p)

        mt = src.discretize_basesource(store)
        ax.scatter(mt.east_shifts, mt.north_shifts, alpha=1)
        ax.scatter(src.east_shift, src.north_shift, color='r')

        plt.axis('equal')
        plt.show()


if __name__ == '__main__':
    util.setup_logging('test_gf_source_types', 'warning')
    unittest.main(defaultTest='GFSourceTypesTestCase')



import os.path as op

from pyrocko import util, moment_tensor
from pyrocko.guts import Timestamp, Float, Int, Bool

from ..base import LocationGenerator

km = 1e3
guts_prefix = 'pf.scenario'


class SourceGenerator(LocationGenerator):

    nevents = Int.T(default=2)
    avoid_water = Bool.T(
        default=False,
        help='Avoid sources offshore under the ocean / lakes.')

    radius = Float.T(
        default=10*km)

    time_min = Timestamp.T(default=util.str_to_time('2017-01-01 00:00:00'))
    time_max = Timestamp.T(default=util.str_to_time('2017-01-03 00:00:00'))

    magnitude_min = Float.T(default=4.0)
    magnitude_max = Float.T(default=0.0)
    b_value = Float.T(
        optional=True, help='Gutenberg Richter magnitude distribution.')

    def __init__(self, *args, **kwargs):
        super(SourceGenerator, self).__init__(*args, **kwargs)
        if self.b_value and self.magnitude_max:
            raise Exception('b_value and magnitude_max are mutually exclusive')

    def draw_magnitude(self, rstate):
        if self.b_value is None:
            return rstate.uniform(self.magnitude_min, self.magnitude_max)
        else:
            return moment_tensor.rand_to_gutenberg_richter(
                rstate.rand(), self.b_value, magnitude_min=self.magnitude_min)

    def get_sources(self):
        sources = []
        for ievent in range(self.nevents):
            src = self.get_source(ievent)
            src.name = 'scenario_ev%03d' % (ievent + 1)
            sources.append(src)

        return sources

    def dump_data(self, path):
        fn_sources = op.join(path, 'sources.yml')
        with open(fn_sources, 'w') as f:
            for src in self.get_sources():
                f.write(src.dump())

        fn_events = op.join(path, 'events.txt')
        with open(fn_events, 'w') as f:
            for isrc, src in enumerate(self.get_sources()):
                f.write(src.pyrocko_event().dump())

        return [fn_events, fn_sources]

    def add_map_artists(self, automap):
        pass



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
from __future__ import absolute_import

from builtins import range
from builtins import map

import os.path as op
import math
from collections import defaultdict

import numpy as num

from pyrocko import util, config
from pyrocko import orthodrome as od
from pyrocko.guts_array import Array
from pyrocko.guts import Object, String


PI = math.pi


class Plate(Object):

    name = String.T(
        help='Name of the tectonic plate.')
    points = Array.T(
        dtype=num.float, shape=(None, 2),
        help='Points on the plate.')

    def max_interpoint_distance(self):
        p = od.latlon_to_xyz(self.points)
        return math.sqrt(num.max(num.sum(
            (p[num.newaxis, :, :] - p[:, num.newaxis, :])**2, axis=2)))

    def contains_point(self, point):
        return od.contains_point(self.points, point)

    def contains_points(self, points):
        return od.contains_points(self.points, points)


class Boundary(Object):

    name1 = String.T()
    name2 = String.T()
    kind = String.T()
    points = Array.T(dtype=num.float, shape=(None, 2))
    cpoints = Array.T(dtype=num.float, shape=(None, 2))
    itypes = Array.T(dtype=num.int, shape=(None))

    def split_types(self, groups=None):
        xyz = od.latlon_to_xyz(self.points)
        xyzmid = (xyz[1:] + xyz[:-1, :]) * 0.5
        cxyz = od.latlon_to_xyz(self.cpoints)
        d = od.distances3d(xyzmid[num.newaxis, :, :], cxyz[:, num.newaxis, :])
        idmin = num.argmin(d, axis=0)
        itypes = self.itypes[idmin]

        if groups is None:
            groupmap = num.arange(len(self._index_to_type))
        else:
            d = {}
            for igroup, group in enumerate(groups):
                for name in group:
                    d[name] = igroup

            groupmap = num.array(
                [d[name] for name in self._index_to_type],
                dtype=num.int)

        iswitch = num.concatenate(
            ([0],
             num.where(groupmap[itypes[1:]] != groupmap[itypes[:-1]])[0]+1,
             [itypes.size]))

        results = []
        for ii in range(iswitch.size-1):
            if groups is not None:
                tt = [self._index_to_type[ityp] for ityp in num.unique(
                    itypes[iswitch[ii]:iswitch[ii+1]])]
            else:
                tt = self._index_to_type[itypes[iswitch[ii]]]

            results.append((tt, self.points[iswitch[ii]:iswitch[ii+1]+1]))

        return results


class Dataset(object):

    def __init__(self, name, data_dir, citation):
        self.name = name
        self.citation = citation
        self.data_dir = data_dir

    def fpath(self, filename):
        return op.join(self.data_dir, filename)

    def download_file(self, url, fpath, username=None, password=None):
        util.download_file(url, fpath, username, password)


class PlatesDataset(Dataset):
    pass


class PeterBird2003(PlatesDataset):
    '''An updated digital model of plate boundaries.'''
    __citation = '''
    Bird, Peter. "An updated digital model of plate boundaries." Geochemistry,
    Geophysics, Geosystems 4.3 (2003).
    '''

    def __init__(
            self,
            name='PeterBird2003',
            data_dir=None,
            raw_data_url=('http://peterbird.name/oldFTP/PB2002/%s')):

        if data_dir is None:
            data_dir = op.join(config.config().tectonics_dir, name)

        PlatesDataset.__init__(
            self,
            name,
            data_dir=data_dir,
            citation=self.__citation)

        self.raw_data_url = raw_data_url

        self.filenames = [
            '2001GC000252_readme.txt',
            'PB2002_boundaries.dig.txt',
            'PB2002_orogens.dig.txt',
            'PB2002_plates.dig.txt',
            'PB2002_poles.dat.txt',
            'PB2002_steps.dat.txt']

        self._full_names = None

    def full_name(self, name):
        if not self._full_names:
            fn = util.data_file(op.join('tectonics', 'bird2003_plates.txt'))

            with open(fn, 'r') as f:
                self._full_names = dict(
                    line.strip().split(None, 1) for line in f)

        return self._full_names[name]

    def download_if_needed(self):
        for fn in self.filenames:
            fpath = self.fpath(fn)
            if not op.exists(fpath):
                self.download_file(self.raw_data_url % fn, fpath)

    def get_boundaries(self):
        self.download_if_needed()
        fpath = self.fpath('PB2002_steps.dat.txt')

        d = defaultdict(list)
        ntyp = 0
        type_to_index = {}
        index_to_type = []
        with open(fpath, 'rb') as f:
            data = []
            for line in f:
                t = line.split()
                s = t[1].lstrip(b':')
                name1 = str(s[0:2].decode('ascii'))
                name2 = str(s[3:5].decode('ascii'))
                kind = s[2]

                alon, alat, blon, blat = list(map(float, t[2:6]))
                mlat = (alat + blat) * 0.5
                dlon = ((blon - alon) + 180.) % 360. - 180.
                mlon = alon + dlon * 0.5
                typ = str(t[14].strip(b':*').decode('ascii'))

                if typ not in type_to_index:
                    ntyp += 1
                    type_to_index[typ] = ntyp - 1
                    index_to_type.append(typ)

                ityp = type_to_index[typ]
                d[name1, kind, name2].append((mlat, mlon, ityp))

        d2 = {}
        for k in d:
            d2[k] = (
                num.array([l[:2] for l in d[k]], dtype=num.float),
                num.array([l[2] for l in d[k]], dtype=num.int))

        fpath = self.fpath('PB2002_boundaries.dig.txt')
        boundaries = []
        name1 = ''
        name2 = ''
        kind = '-'
        with open(fpath, 'rb') as f:
            data = []
            for line in f:
                if line.startswith(b'***'):
                    cpoints, itypes = d2[name1, kind, name2]
                    boundaries.append(Boundary(
                        name1=name1,
                        name2=name2,
                        kind=kind,
                        points=num.array(data, dtype=num.float),
                        cpoints=cpoints,
                        itypes=itypes))

                    boundaries[-1]._type_to_index = type_to_index
                    boundaries[-1]._index_to_type = index_to_type

                    data = []
                elif line.startswith(b' '):
                    data.append(list(map(float, line.split(b',')))[::-1])
                else:
                    s = line.strip()
                    name1 = str(s[0:2].decode('ascii'))
                    name2 = str(s[3:5].decode('ascii'))
                    kind = s[2]

        return boundaries

    def get_plates(self):
        self.download_if_needed()
        fpath = self.fpath('PB2002_plates.dig.txt')
        plates = []
        name = ''
        with open(fpath, 'rb') as f:
            data = []
            for line in f:
                if line.startswith(b'***'):
                    plates.append(Plate(
                        name=name,
                        points=num.array(data, dtype=num.float)))

                    data = []
                elif line.startswith(b' '):
                    data.append(list(map(float, line.split(b',')))[::-1])
                else:
                    name = str(line.strip().decode('ascii'))

        return plates


class StrainRateDataset(Dataset):
    pass


class GSRM1(StrainRateDataset):
    '''Global Strain Rate Map. An integrated global model of present-day
    plate motions and plate boundary deformation'''
    __citation = '''Kreemer, C., W.E. Holt, and A.J. Haines, "An integrated
    global model of present-day plate motions and plate boundary deformation",
    Geophys. J. Int., 154, 8-34, 2003.'''

    def __init__(
            self,
            name='GSRM1.2',
            data_dir=None,
            raw_data_url=('http://gsrm.unavco.org/model/files/1.2/%s')):

        if data_dir is None:
            data_dir = op.join(config.config().tectonics_dir, name)

        StrainRateDataset.__init__(
            self,
            name,
            data_dir=data_dir,
            citation=self.__citation)

        self.raw_data_url = raw_data_url
        self._full_names = None
        self._names = None

    def full_names(self):
        if not self._full_names:
            fn = util.data_file(op.join('tectonics', 'gsrm1_plates.txt'))

            with open(fn, 'r') as f:
                self._full_names = dict(
                    line.strip().split(None, 1) for line in f)

        return self._full_names

    def full_name(self, name):
        name = self.plate_alt_names().get(name, name)
        return self.full_names()[name]

    def plate_names(self):
        if self._names is None:
            self._names = sorted(self.full_names().keys())

        return self._names

    def plate_alt_names(self):
        # 'African Plate' is named 'Nubian Plate' in GSRM1
        return {'AF': 'NU'}

    def download_if_needed(self, fn):
        fpath = self.fpath(fn)
        if not op.exists(fpath):
            self.download_file(self.raw_data_url % fn, fpath)

    def get_velocities(self, reference_name=None, region=None):
        reference_name = self.plate_alt_names().get(
            reference_name, reference_name)

        if reference_name is None:
            reference_name = 'NNR'

        fn = 'velocity_%s.dat' % reference_name

        self.download_if_needed(fn)
        fpath = self.fpath(fn)
        data = []
        with open(fpath, 'rb') as f:
            for line in f:
                if line.strip().startswith(b'#'):
                    continue
                t = line.split()
                data.append(list(map(float, t)))

        arr = num.array(data, dtype=num.float)

        if region is not None:
            points = arr[:, 1::-1]
            mask = od.points_in_region(points, region)
            arr = arr[mask, :]

        lons, lats, veast, vnorth, veast_err, vnorth_err, corr = arr.T
        return lats, lons, vnorth, veast, vnorth_err, veast_err, corr


__all__ = '''
GSRM1
PeterBird2003
Plate'''.split()



# python 2/3
from __future__ import division, print_function, absolute_import
from builtins import zip
from builtins import range
import random
import math
import unittest
import logging
import os.path as op
from tempfile import mkdtemp

from subprocess import check_call
import numpy as num

from pyrocko import util, ahfullgreen, trace, io
from pyrocko.guts import Object, Float, Tuple, List, load

from . import common

guts_prefix = 'test_ahfull'

logger = logging.getLogger('pyrocko.test.test_ahfull')

km = 1000.


def rand(mi, ma):
    return mi + random.random() * (ma-mi)


def g(trs, sta, cha):
    for tr in trs:
        if tr.station == sta and tr.channel == cha:
            return tr


class AhfullKiwiTestSetupEntry(Object):
    vp = Float.T()
    vs = Float.T()
    density = Float.T()
    x = Tuple.T(3, Float.T())
    f = Tuple.T(3, Float.T())
    m6 = Tuple.T(6, Float.T())
    tau = Float.T()
    deltat = Float.T()


class AhfullKiwiTestSetup(Object):
    setups = List.T(AhfullKiwiTestSetupEntry.T())


class AhfullTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    def _make_test_ahfull_kiwi_data(self):
        trs_all = []
        setups = []
        for i in range(100):
            s = AhfullKiwiTestSetupEntry(
                vp=3600.,
                vs=2000.,
                density=2800.,
                x=(rand(100., 1000.), rand(100., 1000.), rand(100., 1000.)),
                f=(rand(-1., 1.), rand(-1., 1.), rand(-1., 1.)),
                m6=tuple(rand(-1., 1.) for _ in range(6)),
                tau=0.005,
                deltat=0.001)

            def dump(stuff, fn):
                with open(fn, 'w') as f:
                    f.write(' '.join('%s' % x for x in stuff))
                    f.write('\n')

            dn = mkdtemp(prefix='test-ahfull-')
            fn_sources = op.join(dn, 'sources.txt')
            fn_receivers = op.join(dn, 'receivers.txt')
            fn_material = op.join(dn, 'material.txt')
            fn_stf = op.join(dn, 'stf.txt')

            dump((0., 0., 0., 0.) + s.m6 + s.f, fn_sources)
            dump(s.x + (1, 1), fn_receivers)
            dump((s.density, s.vp, s.vs), fn_material)

            nstf = int(round(s.tau * 5. / s.deltat))
            t = num.arange(nstf) * s.deltat
            t0 = nstf * s.deltat / 2.
            stf = num.exp(-(t-t0)**2 / (s.tau/math.sqrt(2.))**2)

            stf = num.cumsum(stf)
            stf /= stf[-1]
            stf[0] = 0.0

            data = num.vstack((t, stf)).T
            num.savetxt(fn_stf, data)

            check_call(
                ['ahfull', fn_sources, fn_receivers, fn_material, fn_stf,
                 '%g' % s.deltat, op.join(dn, 'ahfull'), 'mseed', '0'],
                stdout=open('/dev/null', 'w'))

            fns = [op.join(dn, 'ahfull-1-%s-1.mseed' % c) for c in 'xyz']

            trs = []
            for fn in fns:
                trs.extend(io.load(fn))

            for tr in trs:
                tr.set_codes(
                    station='S%03i' % i,
                    channel={'x': 'N', 'y': 'E', 'z': 'D'}[tr.channel])
                tr.shift(-round(t0/tr.deltat)*tr.deltat)

            trs_all.extend(trs)
            setups.append(s)

        setup = AhfullKiwiTestSetup(setups=setups)

        setup.dump(filename=common.test_data_file_no_download(
            'test_ahfull_kiwi_setup.yaml'))
        io.save(trs_all, common.test_data_file_no_download(
            'test_ahfull_kiwi_traces.mseed'))

    def test_ahfull_kiwi(self):
        setup = load(filename=common.test_data_file(
            'test_ahfull_kiwi_setup.yaml'))
        trs_ref = io.load(common.test_data_file(
            'test_ahfull_kiwi_traces.mseed'))

        for i, s in enumerate(setup.setups):
            d3d = math.sqrt(s.x[0]**2 + s.x[1]**2 + s.x[2]**2)

            tlen = d3d / s.vs * 2

            n = int(num.round(tlen / s.deltat))

            out_x = num.zeros(n)
            out_y = num.zeros(n)
            out_z = num.zeros(n)

            ahfullgreen.add_seismogram(
                s.vp, s.vs, s.density, 1000000.0, 1000000.0, s.x, s.f, s.m6,
                'displacement',
                s.deltat, 0.,
                out_x, out_y, out_z,
                ahfullgreen.Gauss(s.tau))

            trs = []
            for out, comp in zip([out_x, out_y, out_z], 'NED'):
                tr = trace.Trace(
                    '', 'S%03i' % i, 'P', comp,
                    deltat=s.deltat, tmin=0.0, ydata=out)

                trs.append(tr)

            trs2 = []

            for cha in 'NED':

                t1 = g(trs, 'S%03i' % i, cha)
                t2 = g(trs_ref, 'S%03i' % i, cha)

                tmin = max(t1.tmin, t2.tmin)
                tmax = min(t1.tmax, t2.tmax)

                t1 = t1.chop(tmin, tmax, inplace=False)
                t2 = t2.chop(tmin, tmax, inplace=False)

                trs2.append(t2)

                d = 2.0 * num.sum((t1.ydata - t2.ydata)**2) / \
                    (num.sum(t1.ydata**2) + num.sum(t2.ydata**2))

                if d >= 0.02:
                    print(d)
                    # trace.snuffle([t1, t2])

                assert d < 0.02

            # trace.snuffle(trs + trs2)


if __name__ == '__main__':
    util.setup_logging('test_ahfull', 'warning')
    unittest.main()



from .base import *  # noqa
from .station import *  # noqa
from .waveform import *  # noqa
from .insar import *  # noqa
from .gnss_campaign import *  # noqa

AVAILABLE_TARGETS =\
    [RandomStationGenerator, WaveformGenerator, InSARGenerator,  # noqa
     GNSSCampaignGenerator]  # noqa



from __future__ import division, print_function, absolute_import

import unittest
import numpy as num
import cProfile
import math
import logging
import shutil

from tempfile import mkdtemp
from .common import Benchmark
from pyrocko import gf, util, cake
from pyrocko.fomosto import qseis, psgrn_pscmp


random = num.random
logger = logging.getLogger('pyrocko.test.test_gf_static')
benchmark = Benchmark()

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.


@unittest.skipUnless(
    qseis.have_backend(), 'backend qseis not available')
@unittest.skipUnless(
    psgrn_pscmp.have_backend(), 'backend psgrn_pscmp not available')
class GFStaticTest(unittest.TestCase):
    tempdirs = []

    def __init__(self, *args, **kwargs):
        self._dummy_store = None
        self.qseis_store_dir = None
        self.pscmp_store_dir = None
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def get_qseis_store_dir(self):
        if self.qseis_store_dir is None:
            self.qseis_store_dir = self._create_qseis_store()

        return self.qseis_store_dir

    def get_pscmp_store_dir(self):
        if self.pscmp_store_dir is None:
            self.pscmp_store_dir = self._create_psgrn_pscmp_store()

        return self.pscmp_store_dir

    def setUp(self):
        return False
        self.cprofile = cProfile.Profile()
        self.cprofile.enable()
        self.addCleanup(
            lambda: self.cprofile.dump_stats(
                '/tmp/process_static_params.prof'))

    def _create_psgrn_pscmp_store(self):
        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
mantle
 35. 8.04 4.48 3.58 1449. 600.
 77.5 8.045 4.49 3.5 1445. 600.
 77.5 8.045 4.49 3.5 180.6 75.
 120. 8.05 4.5 3.427 180. 75.
 120. 8.05 4.5 3.427 182.6 76.06
 165. 8.175 4.509 3.371 188.7 76.55
 210. 8.301 4.518 3.324 201. 79.4
 210. 8.3 4.52 3.321 336.9 133.3
 410. 9.03 4.871 3.504 376.5 146.1
 410. 9.36 5.08 3.929 414.1 162.7
 660. 10.2 5.611 3.918 428.5 172.9
 660. 10.79 5.965 4.229 1349. 549.6'''.lstrip()))

        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)
        store_id = 'psgrn_pscmp_test'
        version = '2008a'

        c = psgrn_pscmp.PsGrnPsCmpConfig()
        c.psgrn_config.sampling_interval = 1.
        c.psgrn_config.version = version
        c.pscmp_config.version = version

        config = gf.meta.ConfigTypeA(
            id=store_id,
            ncomponents=10,
            sample_rate=1./(3600. * 24.),
            receiver_depth=0.*km,
            source_depth_min=0.*km,
            source_depth_max=20.*km,
            source_depth_delta=0.25*km,
            distance_min=0.*km,
            distance_max=40.*km,
            distance_delta=0.25*km,
            modelling_code_id='psgrn_pscmp.%s' % version,
            earthmodel_1d=mod,
            tabulated_phases=[])
        config.validate()

        gf.store.Store.create_editables(
            store_dir, config=config, extra={'psgrn_pscmp': c})

        store = gf.store.Store(store_dir, 'r')
        store.close()

        psgrn_pscmp.build(store_dir, nworkers=4)
        return store_dir

    def _create_qseis_store(self):
        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
mantle
 35. 8.04 4.48 3.58 1449. 600.
 77.5 8.045 4.49 3.5 1445. 600.
 77.5 8.045 4.49 3.5 180.6 75.
 120. 8.05 4.5 3.427 180. 75.
 120. 8.05 4.5 3.427 182.6 76.06
 165. 8.175 4.509 3.371 188.7 76.55
 210. 8.301 4.518 3.324 201. 79.4
 210. 8.3 4.52 3.321 336.9 133.3
 410. 9.03 4.871 3.504 376.5 146.1
 410. 9.36 5.08 3.929 414.1 162.7
 660. 10.2 5.611 3.918 428.5 172.9
 660. 10.79 5.965 4.229 1349. 549.6'''.lstrip()))
        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)

        qsconf = qseis.QSeisConfig()
        qsconf.qseis_version = '2006a'

        qsconf.time_region = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qsconf.cut = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qsconf.wavelet_duration_samples = 0.001
        qsconf.sw_flat_earth_transform = 0

        config = gf.meta.ConfigTypeA(
            id='qseis_test',
            ncomponents=10,
            sample_rate=0.25,
            receiver_depth=0.*km,
            source_depth_min=0*km,
            source_depth_max=10*km,
            source_depth_delta=1*km,
            distance_min=0*km,
            distance_max=20*km,
            distance_delta=2.5*km,
            modelling_code_id='qseis.2006a',
            earthmodel_1d=mod,
            tabulated_phases=[
                gf.meta.TPDef(
                    id='begin',
                    definition='p,P,p\\,P\\'),
                gf.meta.TPDef(
                    id='end',
                    definition='2.5'),
            ])

        config.validate()
        gf.store.Store.create_editables(
            store_dir, config=config, extra={'qseis': qsconf})

        store = gf.store.Store(store_dir, 'r')
        store.make_ttt()
        store.close()

        qseis.build(store_dir, nworkers=1)
        return store_dir

    def test_process_static(self):
        src_length = 5 * km
        src_width = 2 * km
        ntargets = 1600
        interp = ['nearest_neighbor', 'multilinear']
        interpolation = interp[0]

        source = gf.RectangularSource(
            lat=0., lon=0.,
            north_shift=0., east_shift=0., depth=6.5*km,
            width=src_width, length=src_length,
            dip=90., rake=90., strike=90.,
            slip=1.)

        phi = num.zeros(ntargets)              # Horizontal from E
        theta = num.ones(ntargets) * num.pi/2  # Vertical from vertical
        phi.fill(num.deg2rad(192.))
        theta.fill(num.deg2rad(90.-23.))

        sattarget = gf.SatelliteTarget(
            north_shifts=(random.rand(ntargets)-.5) * 25. * km,
            east_shifts=(random.rand(ntargets)-.5) * 25. * km,
            tsnapshot=20,
            interpolation=interpolation,
            phi=phi,
            theta=theta)

        static_target = gf.StaticTarget(
            north_shifts=(random.rand(ntargets)-.5) * 25. * km,
            east_shifts=(random.rand(ntargets)-.5) * 25. * km,
            tsnapshot=20,
            interpolation=interpolation)

        engine = gf.LocalEngine(store_dirs=[self.get_pscmp_store_dir()])

        def process_target(nprocs):

            @benchmark.labeled('process-nearest_neighbor-np%d' % nprocs)
            def process_nearest_neighbor():
                sattarget.interpolation = 'nearest_neighbor'
                return engine.process(source, sattarget, nprocs=nprocs)

            @benchmark.labeled('process-multilinear-np%d' % nprocs)
            def process_multilinear():
                sattarget.interpolation = 'multilinear'
                return engine.process(source, sattarget, nprocs=nprocs)

            return process_nearest_neighbor(), process_multilinear()

        def process_multiple_targets():
            return engine.process(source, [sattarget, static_target])

        for np in [1, 2, 4]:
            nn, ml = process_target(nprocs=np)

        process_multiple_targets()

        # self.plot_static_los_result(ml)
        # print benchmark

    @staticmethod
    def plot_static_los_result(result):
        import matplotlib.pyplot as plt

        fig, _ = plt.subplots(1, 4)

        N = result.request.targets[0].coords5[:, 2]
        E = result.request.targets[0].coords5[:, 3]
        result = result.results_list[0][0].result

        vranges = [(result['displacement.%s' % c].max(),
                    result['displacement.%s' % c].min()) for c in list('ned') +
                   ['los']]

        lmax = num.abs([num.min(vranges), num.max(vranges)]).max()
        levels = num.linspace(-lmax, lmax, 50)

        for dspl, ax in zip(list('ned') + ['los'], fig.axes):
            cmap = ax.tricontourf(E, N, result['displacement.%s' % dspl],
                                  cmap='seismic', levels=levels)
            ax.set_title('displacement.%s' % dspl)
            ax.set_aspect('equal')

        fig.colorbar(cmap)
        plt.show()

    def test_sum_static(self):
        from pyrocko.gf import store_ext
        benchmark.show_factor = True

        store = gf.Store(self.get_qseis_store_dir())
        store.open()
        src_length = 2 * km
        src_width = 5 * km
        ntargets = 1600
        interp = ['nearest_neighbor', 'multilinear']
        interpolation = interp[0]

        source = gf.RectangularSource(
            lat=0., lon=0.,
            depth=5*km, north_shift=0., east_shift=0.,
            width=src_width, length=src_length)

        static_target = gf.StaticTarget(
            north_shifts=5*km + random.rand(ntargets) * 5*km,
            east_shifts=0*km + random.rand(ntargets) * 5*km)
        targets = static_target.get_targets()

        dsource = source.discretize_basesource(store, targets[0])
        source_coords_arr = dsource.coords5()
        mts_arr = dsource.m6s

        receiver_coords_arr = num.empty((len(targets), 5))
        for itarget, target in enumerate(targets):
            receiver = target.receiver(store)
            receiver_coords_arr[itarget, :] = \
                [receiver.lat, receiver.lon, receiver.north_shift,
                 receiver.east_shift, receiver.depth]

        def sum_target(cstore, irecords, delays_t, delays_s,
                       weights, pos, nthreads):

            @benchmark.labeled('sum-timeseries-np%d' % nthreads)
            def sum_timeseries():
                nsummands = weights.size // ntargets
                res = num.zeros(ntargets)
                for t in range(ntargets):
                    sl = slice(t*nsummands, (t+1) * nsummands)
                    r = store_ext.store_sum(
                        cstore, irecords[sl], delays_t[sl],
                        weights[sl], pos, 1)
                    res[t] = r[0]
                return res

            @benchmark.labeled('sum-static-np%d' % nthreads)
            def sum_static():
                return store_ext.store_sum_static(
                    cstore, irecords, delays_s, weights,
                    pos, ntargets, nthreads)

            return sum_timeseries(), sum_static()

        args = (store.cstore, source_coords_arr, mts_arr, receiver_coords_arr,
                'elastic10', interpolation, 0)

        benchmark.clear()
        for nthreads in [1, 2, 4]:
            for (weights, irecords) in store_ext.make_sum_params(*args):
                delays_t = num.zeros_like(weights, dtype=num.float64)
                delays_s = dsource.times.astype(num.float64)
                pos = 6
                t, s = sum_target(store.cstore, irecords, delays_t, delays_s,
                                  weights, pos, nthreads)
                # print benchmark.__str__(header=False)
                num.testing.assert_equal(t, s)
                benchmark.clear()

    def test_gnss_target(self):
        src_length = 5 * km
        src_width = 2 * km
        nstations = 100
        interp = ['nearest_neighbor', 'multilinear']
        interpolation = interp[0]

        source = gf.RectangularSource(
            lat=0., lon=0.,
            north_shift=0., east_shift=0., depth=6.5*km,
            width=src_width, length=src_length,
            dip=90., rake=90., strike=90.,
            slip=1.)

        gnss_target = gf.GNSSCampaignTarget(
            lats=(random.uniform(-.2, .2, nstations)),
            lons=(random.uniform(-.2, .2, nstations)),
            interpolation=interpolation)

        engine = gf.LocalEngine(store_dirs=[self.get_pscmp_store_dir()])
        res = engine.process(source, gnss_target, nprocs=0)

        statics = res.static_results()
        for static in statics:
            assert len(static.campaign.stations) == nstations

    def test_new_static(self):
        from pyrocko.gf import store_ext
        benchmark.show_factor = True

        store = gf.Store(self.get_pscmp_store_dir())
        store.open()
        src_length = 2 * km
        src_width = 2 * km
        ntargets = 20

        north_shifts, east_shifts = num.meshgrid(
            num.linspace(-20*km, 20*km, ntargets),
            num.linspace(-20*km, 20*km, ntargets))

        interp = ['nearest_neighbor', 'multilinear']
        interpolation = interp[1]

        source = gf.RectangularSource(
            lat=0., lon=0.,
            depth=5*km, north_shift=0., east_shift=0.,
            anchor='top',
            width=src_width, length=src_length)

        static_target = gf.GNSSCampaignTarget(
            north_shifts=north_shifts,
            east_shifts=east_shifts,
            lats=num.zeros_like(north_shifts),
            lons=num.zeros_like(north_shifts))

        targets = static_target.get_targets()

        dsource = source.discretize_basesource(store, targets[0])
        mts_arr = dsource.m6s
        delays_s = dsource.times.astype(num.float64)
        pos = 1

        scheme_desc = ['displacement.n', 'displacement.e', 'displacement.d']

        benchmark.clear()

        def run(interpolation=interp[0], nthreads=1, niter=1):

            @benchmark.labeled(' sum_statics %d cpu (%s)' %
                               (nthreads, interpolation))
            def fwd_model_seperate(interpolation=interp[0]):
                args = (store.cstore, dsource.coords5(), mts_arr,
                        static_target.coords5, 'elastic10', interpolation,
                        nthreads)

                sum_params = store_ext.make_sum_params(*args)

                out = {}

                for icomp, comp in enumerate(scheme_desc):
                    weights, irecords = sum_params[icomp]
                    out[comp] = store_ext.store_sum_static(
                            store.cstore, irecords, delays_s, weights,
                            pos, ntargets**2, nthreads)
                return out

            @benchmark.labeled('calc_statics %d cpu (%s)' %
                               (nthreads, interpolation))
            def fwd_model_unified(interpolation=interp[0]):
                out = {}
                res = store_ext.store_calc_static(
                        store.cstore,
                        dsource.coords5(),
                        mts_arr,
                        dsource.times,
                        static_target.coords5,
                        'elastic10',
                        interpolation,
                        pos,
                        nthreads)
                for comp, r in zip(scheme_desc, res):
                    out[comp] = r

                return out

            for _ in range(niter):
                res1 = fwd_model_seperate(interpolation)
            for _ in range(niter):
                res2 = fwd_model_unified(interpolation)

            for r1, r2 in zip(res1.values(), res2.values()):
                num.testing.assert_equal(r1, r2)

        for interpolation in interp:
            continue
            for nthreads in [1, 2, 4, 8, 0]:
                run(interpolation, nthreads)
            print(benchmark)
            benchmark.clear()

        run(interpolation, nthreads=0, niter=30)
        print(benchmark)

        def plot(displ):
            import matplotlib.pyplot as plt
            size = int(num.sqrt(displ.size))
            fig = plt.figure()
            ax = fig.gca()
            ax.imshow(displ.reshape((size, size)))
            plt.show()

        # plot(res1['displacement.n'])


if __name__ == '__main__':
    util.setup_logging('test_gf', 'warning')
    unittest.main(defaultTest='GFStaticTest')



from __future__ import division, print_function, absolute_import
from builtins import zip
import random
import unittest
import multiprocessing
import time
from collections import defaultdict
import numpy as num
from . import common

from pyrocko import util, trace, autopick

from pyrocko.parstack import parstack, get_offset_and_length
from pyrocko.parstack import argmax as pargmax


def numeq(a, b, eps):
    return (num.all(num.asarray(a).shape == num.asarray(b).shape and
            num.abs(num.asarray(a) - num.asarray(b)) < eps))


def g(l, a):
    return num.array([getattr(x, a) for x in l])


class ParstackTestCase(unittest.TestCase):

    def test_parstack(self):
        for i in range(100):
            narrays = random.randint(1, 5)
            arrays = [
                num.random.random(random.randint(5, 10))
                for j in range(narrays)
            ]
            offsets = num.random.randint(-5, 6, size=narrays).astype(num.int32)
            nshifts = random.randint(1, 10)
            shifts = num.random.randint(
                -5, 6, size=(nshifts, narrays)).astype(num.int32)
            weights = num.random.random((nshifts, narrays))

            for method in (0, 1):
                for nparallel in range(1, 5):
                    r1, o1 = parstack(
                        arrays, offsets, shifts, weights, method,
                        impl='openmp',
                        nparallel=nparallel)

                    r2, o2 = parstack(
                        arrays, offsets, shifts, weights, method, impl='numpy')

                    assert o1 == o2
                    assert numeq(r1, r2, 1e-9)

    def test_parstack_limited(self):
        for i in range(10):
            narrays = random.randint(1, 5)
            arrays = [
                num.random.random(random.randint(5, 10))
                for j in range(narrays)
            ]
            offsets = num.random.randint(-5, 6, size=narrays).astype(num.int32)
            nshifts = random.randint(1, 10)
            shifts = num.random.randint(
                -5, 6, size=(nshifts, narrays)).astype(num.int32)
            weights = num.random.random((nshifts, narrays))

            for nparallel in range(1, 5):
                r1, o1 = parstack(
                    arrays, offsets, shifts, weights, 0,
                    nparallel=nparallel,
                    impl='openmp')

                for impl in ['openmp', 'numpy']:
                    r2, o2 = parstack(
                        arrays, offsets, shifts, weights, 0,
                        lengthout=r1.shape[1],
                        offsetout=o1,
                        nparallel=nparallel,
                        impl=impl)

                    assert o1 == o2
                    num.testing.assert_almost_equal(
                        r1, r2, decimal=9)

                    n = r1.shape[1]
                    for k in range(n):
                        r3, o3 = parstack(
                            arrays, offsets, shifts, weights, 0,
                            lengthout=n,
                            offsetout=o1-k,
                            nparallel=nparallel,
                            impl=impl)

                        assert o3 == o1-k
                        num.testing.assert_almost_equal(
                            r1[:, :n-k], r3[:, k:], decimal=9)

                    for k in range(n):
                        r3, o3 = parstack(
                            arrays, offsets, shifts, weights, 0,
                            lengthout=n,
                            offsetout=o1+k,
                            nparallel=nparallel,
                            impl=impl)

                        assert o3 == o1+k
                        num.testing.assert_almost_equal(
                            r1[:, k:], r3[:, :n-k], decimal=9)

                    for k in range(n):
                        r3, o3 = parstack(
                            arrays, offsets, shifts, weights, 0,
                            lengthout=n-k,
                            offsetout=o1,
                            nparallel=nparallel,
                            impl=impl)

                        assert o3 == o1
                        num.testing.assert_almost_equal(
                            r1[:, :n-k], r3[:, :], decimal=9)
                        # assert numeq(r1[:, :n-k], r3[:, :], 1e-9)

    def test_parstack_cumulative(self):
        for i in range(10):
            narrays = random.randint(1, 5)
            arrays = [
                num.random.random(random.randint(5, 10))
                for i in range(narrays)
            ]
            offsets = num.random.randint(-5, 6, size=narrays).astype(num.int32)
            nshifts = random.randint(1, 10)
            shifts = num.random.randint(
                -5, 6, size=(nshifts, narrays)).astype(num.int32)
            weights = num.random.random((nshifts, narrays))

            for method in (0,):
                for nparallel in range(1, 4):
                    result, offset = parstack(
                        arrays, offsets, shifts, weights, method,
                        result=None,
                        nparallel=nparallel,
                        impl='openmp')

                    result1 = result.copy()
                    for k in range(5):
                        result, offset = parstack(
                            arrays, offsets, shifts, weights, method,
                            result=result,
                            nparallel=nparallel,
                            impl='openmp')

                        num.testing.assert_almost_equal(
                            result, result1*(k+2.), decimal=9)

    def benchmark(self):

        for nsamples in (10, 100, 1000, 10000):
            nrepeats = max(10, 1000 / nsamples)

            narrays = 20
            arrays = []
            for iarray in range(narrays):
                arrays.append(num.arange(nsamples, dtype=num.float))

            offsets = num.arange(narrays, dtype=num.int32)

            nshifts = 100
            shifts = num.zeros((nshifts, narrays), dtype=num.int32)
            weights = num.ones((nshifts, narrays))

            confs = [('numpy', 1)]
            for nparallel in range(1, multiprocessing.cpu_count() + 1):
                confs.append(('openmp', nparallel))

            for (impl, nparallel) in confs:
                t0 = time.time()
                for j in range(nrepeats):
                    r, o = parstack(
                        arrays, offsets, shifts, weights, 0,
                        impl=impl, nparallel=nparallel)

                t1 = time.time()

                t = t1-t0
                score = nsamples * narrays * nshifts * nrepeats / t / 1e9
                print('%s, %i, %i, %g' % (impl, nparallel, nsamples, score))

    @unittest.skip('needs manual inspection')
    @common.require_gui
    def _off_test_synthetic(self):

        from pyrocko import gf

        km = 1000.
        nstations = 10
        edepth = 5*km
        store_id = 'crust2_d0'

        swin = 2.
        lwin = 9.*swin
        ks = 1.0
        kl = 1.0
        kd = 3.0

        engine = gf.get_engine()
        snorths = (num.random.random(nstations)-1.0) * 50*km
        seasts = (num.random.random(nstations)-1.0) * 50*km
        targets = []
        for istation, (snorths, seasts) in enumerate(zip(snorths, seasts)):
            targets.append(
                gf.Target(
                    quantity='displacement',
                    codes=('', 's%03i' % istation, '', 'Z'),
                    north_shift=float(snorths),
                    east_shift=float(seasts),
                    store_id=store_id,
                    interpolation='multilinear'))

        source = gf.DCSource(
            north_shift=50*km,
            east_shift=50*km,
            depth=edepth)

        store = engine.get_store(store_id)

        response = engine.process(source, targets)
        trs = []

        station_traces = defaultdict(list)
        station_targets = defaultdict(list)
        for source, target, tr in response.iter_results():
            tp = store.t('any_P', source, target)
            t = tp - 5 * tr.deltat + num.arange(11) * tr.deltat
            if False:
                gauss = trace.Trace(
                    tmin=t[0],
                    deltat=tr.deltat,
                    ydata=num.exp(-((t-tp)**2)/((2*tr.deltat)**2)))

                tr.ydata[:] = 0.0
                tr.add(gauss)

            trs.append(tr)
            station_traces[target.codes[:3]].append(tr)
            station_targets[target.codes[:3]].append(target)

        station_stalta_traces = {}
        for nsl, traces in station_traces.items():
            etr = None
            for tr in traces:
                sqr_tr = tr.copy(data=False)
                sqr_tr.ydata = tr.ydata**2
                if etr is None:
                    etr = sqr_tr
                else:
                    etr += sqr_tr

            autopick.recursive_stalta(swin, lwin, ks, kl, kd, etr)
            etr.set_codes(channel='C')

            station_stalta_traces[nsl] = etr

        trace.snuffle(trs + list(station_stalta_traces.values()))
        deltat = trs[0].deltat

        nnorth = 50
        neast = 50

        size = 200*km

        north = num.linspace(-size, size, nnorth)
        north2 = num.repeat(north, neast)
        east = num.linspace(-size, size, neast)
        east2 = num.tile(east, nnorth)
        depth = 5*km

        def tcal(target, i):
            try:
                return store.t(
                    'any_P',
                    gf.Location(
                        north_shift=north2[i],
                        east_shift=east2[i],
                        depth=depth),
                    target)

            except gf.OutOfBounds:
                return 0.0

        nsls = sorted(station_stalta_traces.keys())

        tts = num.fromiter((tcal(station_targets[nsl][0], i)
                           for i in range(nnorth*neast)
                           for nsl in nsls), dtype=num.float)

        arrays = [
            station_stalta_traces[nsl].ydata.astype(num.float) for nsl in nsls]
        offsets = num.array(
            [int(round(station_stalta_traces[nsl].tmin / deltat))
             for nsl in nsls], dtype=num.int32)
        shifts = -num.array(
            [int(round(tt / deltat))
             for tt in tts], dtype=num.int32).reshape(nnorth*neast, nstations)
        weights = num.ones((nnorth*neast, nstations))

        print(shifts[25*neast + 25] * deltat)

        print(offsets.dtype, shifts.dtype, weights.dtype)

        print('stack start')
        mat, ioff = parstack(arrays, offsets, shifts, weights, 1)
        print('stack stop')

        mat = num.reshape(mat, (nnorth, neast))

        from matplotlib import pyplot as plt

        fig = plt.figure()

        axes = fig.add_subplot(1, 1, 1, aspect=1.0)

        axes.contourf(east/km, north/km, mat)

        axes.plot(
            g(targets, 'east_shift')/km,
            g(targets, 'north_shift')/km, '^')
        axes.plot(source.east_shift/km, source.north_shift/km, 'o')
        plt.show()

    def test_argmax(self):
        a = num.random.random((100, 1000))
        argmax_numpy = num.argmax(a, axis=0)
        nparallel = 4
        argmax_parstack = pargmax(a, nparallel)

        num.testing.assert_almost_equal(
            argmax_parstack.astype(num.int64), argmax_numpy)

    def test_limited(self):
        arrays = [
            num.array([0, 0, 0, 0, 0, 0, 0, 1, 0, -1, 0], dtype=num.float),
            num.array([0, 0, 0, 0, 0, 0, 1, 0, -1, 0, 0], dtype=num.float),
            num.array([0, 0, 0, 0, 0, 1, 0, -1, 0, 0, 0], dtype=num.float)]

        offsets = num.array([0, 0, 0], dtype=num.int32)
        shifts = -num.array([
            [8, 7, 6],
            [7, 6, 5],
            [6, 5, 4]], dtype=num.int32)

        weights = num.ones((3, 3), dtype=num.float)

        mat, ioff = parstack(arrays, offsets, shifts, weights, 0)

        ioff_total, nsamples_total = get_offset_and_length(
            arrays, offsets, shifts)

        mat0, ioff = parstack(arrays, offsets, shifts, weights, 0)

        neach = 3
        for ioff in range(0, nsamples_total, 3):
            mat, ioff_check = parstack(
                arrays, offsets, shifts, weights, 0,
                offsetout=ioff_total + ioff,
                lengthout=neach)

            assert ioff_total + ioff == ioff_check

            num.testing.assert_almost_equal(
                mat0[:, ioff:ioff+neach], mat)


if __name__ == '__main__':
    util.setup_logging('test_parstack', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import unittest
import shutil
import tempfile
from pyrocko import util, gf
from pyrocko.fomosto import qssp


@unittest.skipUnless(
    qssp.have_backend(), 'backend qssp not available')
class QSSPTestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='pyrocko.qseis')

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def test_qssp_build(self):
        qssp.init(self.tmpdir, '2010')
        store = gf.store.Store(self.tmpdir, 'r')
        store.make_ttt()
        qssp.build(self.tmpdir)


if __name__ == '__main__':
    util.setup_logging('test_gf_qssp', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import range, zip

import unittest
import numpy as num
import cProfile
import math
import logging
import shutil

from tempfile import mkdtemp
from .common import Benchmark
from pyrocko import gf, util

random = num.random
logger = logging.getLogger('pyrocko.test.test_gf_benchmark')
benchmark = Benchmark()

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.


class GFBenchmarkTest(unittest.TestCase):

    tempdirs = []

    def __init__(self, *args, **kwargs):
        self._dummy_store = None
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def dummy_store(self):
        if self._dummy_store is None:

            conf = gf.ConfigTypeA(
                id='empty_regional',
                source_depth_min=0.,
                source_depth_max=20*km,
                source_depth_delta=1*km,
                distance_min=1*km,
                distance_max=2000*km,
                distance_delta=1*km,
                sample_rate=2.0,
                ncomponents=10)

            store_dir = mkdtemp(prefix='gfstore')
            self.tempdirs.append(store_dir)

            gf.Store.create(store_dir, config=conf)
            self._dummy_store = gf.Store(store_dir)

        return self._dummy_store

    def setUp(self):
        self.cprofile = cProfile.Profile()
        self.cprofile.enable()
        self.addCleanup(
            lambda: self.cprofile.dump_stats('/tmp/make_sum_params.cprof'))

    def _test_sum_benchmark(self):
        from pyrocko.gf import store_ext
        benchmark.show_factor = True

        def test_weights_bench(store, dim, ntargets, interpolation):
            source = gf.RectangularSource(
                lat=1.4, lon=-1.6,
                depth=10*km, north_shift=0.1, east_shift=0.1,
                width=dim, length=dim)

            targets = [gf.Target(
                lat=random.random()*10.,
                lon=random.random()*10.,
                north_shift=0.1,
                east_shift=0.1) for x in range(ntargets)]

            dsource = source.discretize_basesource(store, targets[0])
            source_coords_arr = dsource.coords5()
            mts_arr = dsource.m6s

            receiver_coords_arr = num.empty((len(targets), 5))
            for itarget, target in enumerate(targets):
                receiver = target.receiver(store)
                receiver_coords_arr[itarget, :] = \
                    [receiver.lat, receiver.lon, receiver.north_shift,
                     receiver.east_shift, receiver.depth]
            ns = mts_arr.shape[0]
            label = '_ns%04d_nt%04d_%s' % (ns,
                                           len(targets),
                                           interpolation)

            @benchmark.labeled('c%s' % label)
            def sum_c():
                return store_ext.make_sum_params(
                    store.cstore,
                    source_coords_arr,
                    mts_arr,
                    receiver_coords_arr,
                    'elastic10',
                    interpolation,
                    0)

            @benchmark.labeled('p%s' % label)
            def sum_python():
                weights_c = []
                irecords_c = []
                for itar, target in enumerate(targets):
                    receiver = target.receiver(store)
                    dsource = source.discretize_basesource(store, target)

                    for i, (component, args, delays, weights) in \
                            enumerate(store.config.make_sum_params(
                                dsource, receiver)):
                        if len(weights_c) <= i:
                            weights_c.append([])
                            irecords_c.append([])

                        if interpolation == 'nearest_neighbor':
                            irecords = num.array(store.config.irecords(*args))
                            weights = num.array(weights)
                        else:
                            assert interpolation == 'multilinear'
                            irecords, ip_weights =\
                                store.config.vicinities(*args)
                            neach = irecords.size // args[0].size
                            weights = num.repeat(weights, neach) * ip_weights
                            delays = num.repeat(delays, neach)

                        weights_c[i].append(weights)
                        irecords_c[i].append(irecords)
                for c in range(len(weights_c)):
                    weights_c[c] = num.concatenate([w for w in weights_c[c]])
                    irecords_c[c] = num.concatenate([ir for
                                                     ir in irecords_c[c]])

                return list(zip(weights_c, irecords_c))

            rc = sum_c()
            rp = sum_python()

            print(benchmark.__str__(header=False))
            benchmark.clear()

            # Comparing the results
            if isinstance(store.config, gf.meta.ConfigTypeA):
                idim = 4
            elif isinstance(store.config, gf.meta.ConfigTypeB):
                idim = 8
            if interpolation == 'nearest_neighbor':
                idim = 1

            for i, nsummands in enumerate([6, 6, 4]):
                for r in [0, 1]:
                    r_c = rc[i][r]
                    r_p = rp[i][r].reshape(ntargets, nsummands, ns*idim)
                    r_p = num.transpose(r_p, axes=[0, 2, 1])

                    num.testing.assert_almost_equal(r_c, r_p.flatten())
                if False:
                    print('irecord_c: {0:>7}, {1:>7}'.format(
                        rc[i][1].min(), rc[i][1].max()))
                    print('irecord_p: {0:>7}, {1:>7}'.format(
                        rp[i][1].min(), rp[i][1].max()))

                if False:
                    print('weights_c: {0:>7}, {1:>7}'.format(
                        rc[i][0].min(), rc[i][0].max()))
                    print('weights_p: {0:>7}, {1:>7}'.format(
                        rp[i][0].min(), rp[i][0].max()))

        '''
        Testing loop
        '''
        dims = [2*km, 5*km, 8*km, 16*km]
        ntargets = [10, 100, 1000]

        dims = [16*km]
        ntargets = [1000]
        store = self.dummy_store()
        store.open()

        for interpolation in ['multilinear', 'nearest_neighbor']:
            for d in dims:
                for nt in ntargets:
                    test_weights_bench(store, d, nt, interpolation)


if __name__ == '__main__':
    util.setup_logging('test_gf', 'warning')
    unittest.main(defaultTest='GFBenchmarkTest.test_sum_benchmark')



from __future__ import division, print_function, absolute_import
from builtins import range
from pyrocko import trace, pile, io, config, util

import unittest
import numpy as num
import tempfile
import random
import os
from random import choice as rc
from os.path import join as pjoin


def numeq(a, b, eps):
    return num.all(num.abs(num.array(a) - num.array(b)) < eps)


def makeManyFiles(nfiles, nsamples, networks, stations, channels, tmin):

    datadir = tempfile.mkdtemp()
    traces = []
    deltat = 1.0
    for i in range(nfiles):
        ctmin = tmin+i*nsamples*deltat  # random.randint(1,int(time.time()))

        data = num.ones(nsamples)
        traces.append(
            trace.Trace(
                rc(networks), rc(stations), '', rc(channels),
                ctmin, None, deltat, data))

    fnt = pjoin(
        datadir,
        '%(network)s-%(station)s-%(location)s-%(channel)s-%(tmin)s.mseed')

    io.save(traces, fnt, format='mseed')

    return datadir


class PileTestCase(unittest.TestCase):

    def testPileTraversal(self):
        import shutil
        config.show_progress = False
        nfiles = 200
        nsamples = 1000

        abc = 'abcdefghijklmnopqrstuvwxyz'

        def rn(n):
            return ''.join([random.choice(abc) for i in range(n)])

        stations = [rn(4) for i in range(10)]
        channels = [rn(3) for i in range(3)]
        networks = ['xx']

        tmin = 1234567890
        datadir = makeManyFiles(
            nfiles, nsamples, networks, stations, channels, tmin)
        filenames = util.select_files([datadir], show_progress=False)
        cachedir = pjoin(datadir, '_cache_')
        p = pile.Pile()
        p.load_files(filenames=filenames, cache=pile.get_cache(cachedir),
                     show_progress=False)

        assert set(p.networks) == set(networks)
        assert set(p.stations) == set(stations)
        assert set(p.channels) == set(channels)

        ntr = 0
        for tr in p.iter_traces():
            ntr += 1
            assert tr.data_len() == nsamples

        assert ntr == nfiles

        toff = 0
        while toff < nfiles*nsamples:

            trs, loaded1 = p.chop(tmin+10, tmin+200)
            for tr in trs:
                assert num.all(tr.get_ydata() == num.ones(190))

            trs, loaded2 = p.chop(tmin-100, tmin+100)
            for tr in trs:
                assert len(tr.get_ydata()) == 100

            loaded = loaded1 | loaded2
            while loaded:
                file = loaded.pop()
                file.drop_data()

            toff += nsamples

        s = 0
        for traces in p.chopper(tmin=None, tmax=p.tmax+1., tinc=122.,
                                degap=False):
            for tr in traces:
                s += num.sum(tr.ydata)

        assert int(round(s)) == nfiles*nsamples

        for fn in filenames:
            os.utime(fn, None)

        p.reload_modified()

        pile.get_cache(cachedir).clean()
        shutil.rmtree(datadir)

    def testMemTracesFile(self):
        tr = trace.Trace(ydata=num.arange(100, dtype=num.float))

        f = pile.MemTracesFile(None, [tr])
        p = pile.Pile()
        p.add_file(f)
        for tr in p.iter_all(include_last=True):
            assert numeq(tr.ydata, num.arange(100, dtype=num.float), 0.001)


if __name__ == "__main__":
    util.setup_logging('test_pile', 'warning')
    unittest.main()



# python 2/3
from __future__ import division, print_function, absolute_import
from builtins import range

import math
import unittest
import numpy as num
from io import BytesIO

from pyrocko import cake, util

km = 1000.


class CakeTestCase(unittest.TestCase):

    def test_copy(self):
        m1 = cake.Material(vp=1., vs=1.)
        m2 = cake.Material(vp=4., vs=4.)

        s = cake.Surface(0., m1)
        self.assertEqual(s.__dict__, s.copy().__dict__)

        i = cake.Interface(1., m1, m2, name='a')
        self.assertEqual(i.__dict__, i.copy().__dict__)

    def test_layer_resize(self):
        depth_min_init = 0.
        depth_max_init = 3.
        m1 = cake.Material(vp=1., vs=1.)
        m2 = cake.Material(vp=4., vs=4.)

        tests = [
            (cake.HomogeneousLayer, 1., None, 2., m1, m1),
            (cake.HomogeneousLayer, None, 2., 1., m1, m1),
            (cake.HomogeneousLayer, 1., 2., 1.5, m1, m1),
            (cake.GradientLayer, 1., None, 2., cake.Material(2., 2.), m2),
            (cake.GradientLayer, None, 2., 1., m1, cake.Material(3., 3.)),
            (cake.GradientLayer, 1., 2., 1.5,
                cake.Material(2., 2.,), cake.Material(3., 3.)),
            (cake.GradientLayer, None, 4., 2., m1, cake.Material(5., 5.)),
        ]

        for t in tests:
            cls, depth_min, depth_max, depth_mid, nmtop, nmbot = t
            if cls == cake.HomogeneousLayer:
                layer = cls(ztop=depth_min_init, zbot=depth_max_init, m=m1)
            if cls == cake.GradientLayer:
                layer = cls(ztop=depth_min_init, zbot=depth_max_init,
                            mtop=m1, mbot=m2)
            depth_min = depth_min or depth_min_init
            depth_max = depth_max or depth_max_init

            layer.resize(depth_min=depth_min, depth_max=depth_max)
            self.assertEqual(layer.ztop, depth_min)
            self.assertEqual(layer.zbot, depth_max)
            self.assertEqual(layer.zmid, num.mean((depth_min, depth_max)))

            self.assertEqual(layer.mtop, nmtop)
            self.assertEqual(layer.mbot, nmbot)

    def test_single_layer_extract(self):
        mod = cake.load_model()
        zmin = 100.
        zmax = 200.
        new_mod = mod.extract(zmin, zmax)
        elements = list(new_mod.elements())
        interface_material_top = mod.material(zmin)
        interface_material_bot = mod.material(zmax)
        if not isinstance(elements[0], cake.Surface):
            self.assertEqual(elements[0].mtop, interface_material_top)
        self.assertEqual(elements[-1].mbot, interface_material_bot)
        self.assertEqual(elements[0].ztop, zmin)
        self.assertEqual(elements[-1].zbot, zmax)

    def test_random_model_extract(self):
        nz = 100
        mod = cake.load_model()
        layers = list(mod.elements())
        zmin = layers[0].ztop
        zmax = layers[mod.nlayers-1].zbot
        zmins = num.random.uniform(zmin, zmax, nz)
        zmaxs = num.random.uniform(zmin, zmax, nz)
        for i in range(nz):
            zmin = min(zmins[i], zmaxs[i])
            zmax = max(zmins[i], zmaxs[i])
            new_mod = mod.extract(zmin, zmax)

            elements = list(new_mod.elements())
            n_layers = len([e for e in elements if isinstance(e, cake.Layer)])
            interface_material_top = mod.material(zmin)
            interface_material_bot = mod.material(zmax)
            if not isinstance(elements[0], cake.Surface):
                self.assertEqual(elements[0].mtop, interface_material_top)
            if isinstance(elements[-1], cake.Layer):
                self.assertEqual(elements[-1].ilayer, n_layers-1)
                for k, v in elements[-1].mbot.__dict__.items():
                    self.assertAlmostEqual(
                        v, interface_material_bot.__dict__[k], 6)
            self.assertEqual(elements[0].ztop, zmin)
            self.assertEqual(elements[-1].zbot, zmax)

    def test_interface_model_extract(self):
        nz = 100
        mod = cake.load_model()
        layers = list(mod.elements())
        for i in range(nz):
            i = num.random.randint(0, len(layers)-3)
            i2 = num.random.randint(i+2, len(layers)-1)
            z1 = layers[i].zbot
            z2 = layers[i2].zbot
            zmin = min(z1, z2)
            zmax = max(z1, z2)
            new_mod = mod.extract(zmin, zmax)
            interface_material_top = mod.material(zmin)
            interface_material_bot = mod.material(zmax)
            elements = list(new_mod.elements())
            n_layers = len([e for e in elements if isinstance(e, cake.Layer)])
            if isinstance(elements[0], cake.Layer):
                self.assertEqual(elements[0].mtop, interface_material_top)
                self.assertEqual(elements[0].ilayer, 0)
            if isinstance(elements[-1], cake.Layer):
                self.assertEqual(elements[-1].ilayer, n_layers-1)
                self.assertEqual(elements[-1].mbot, interface_material_bot)
            self.assertEqual(elements[0].ztop, zmin)
            self.assertEqual(elements[-1].zbot, zmax)
            if zmin == 0.:
                assert isinstance(elements[0], cake.Surface)

    def test_material(self):
        mat = cake.Material(
            poisson=0.20, rho=3000., qp=100.)
        mat1 = cake.Material(
            vp=mat.vp, poisson=mat.poisson(), rho=mat.rho, qp=mat.qp)
        mat2 = cake.Material(
            vp=mat.vp, vs=mat1.vs, rho=mat1.rho, qs=mat1.qs)
        mat3 = cake.Material(
            lame=mat2.lame(), rho=mat2.rho, qp=mat2.qp, qs=mat2.qs)
        mat4 = cake.Material(
            vs=mat3.vs, poisson=mat3.poisson(), rho=mat3.rho,
            qk=mat3.qk(), qmu=mat3.qmu())

        mat5 = eval('cake.'+repr(mat))

        for matx in (mat1, mat2, mat3, mat4, mat5):
            self.assertEqual(mat.vp, matx.vp)
            self.assertEqual(mat.vs, matx.vs)
            self.assertEqual(mat.rho, matx.rho)
            self.assertEqual(mat.qp, matx.qp)
            self.assertEqual(mat.qs, matx.qs)
            self.assertEqual(mat.describe(), matx.describe())

    def test_classic(self):
        phase = cake.PhaseDef.classic('PP')[0]
        assert str(phase) == '''
Phase definition "P<(cmb)(moho)pP<(cmb)(moho)p":
 - P mode propagation, departing downward \
(may not propagate deeper than interface cmb)
 - passing through moho on upgoing path
 - P mode propagation, departing upward
 - surface reflection
 - P mode propagation, departing downward \
(may not propagate deeper than interface cmb)
 - passing through moho on upgoing path
 - P mode propagation, departing upward
 - arriving at target from below'''.strip()

        mod = cake.load_model()
        rays = mod.arrivals(
            phases=[phase], distances=[5000*km*cake.m2d], zstart=500.)

        assert str(rays[0]).split() == '''10.669 s/deg    5000 km  601.9 s  \
33.8   33.8  17%  12% P<(cmb)(moho)pP<(cmb)(moho)p (P^0P)            \
0_1_2_3_(4-5)_(6-7)_8_(7-6)_(5-4)_3_2_1_0|\
0_1_2_3_(4-5)_(6-7)_8_(7-6)_(5-4)_3_2_1_0'''.split()

        assert abs(rays[0].t - 601.9) < 0.2

    def test_simplify(self):
        mod1 = cake.load_model()
        mod2 = mod1.simplify()
        phases = cake.PhaseDef.classic('Pdiff')
        rays1 = mod1.arrivals(phases=phases, distances=[120.], zstart=500.)
        rays2 = mod2.arrivals(phases=phases, distances=[120.], zstart=500.)
        assert abs(rays1[0].t - 915.9) < 0.1
        assert abs(rays2[0].t - 915.9) < 0.1

    def test_path(self):
        mod = cake.load_model()
        phase = cake.PhaseDef('P')
        ray = mod.arrivals(phases=[phase], distances=[70.], zstart=100.)

        z, x, t = ray[0].zxt_path_subdivided()
        assert z[0].size == 681

    def test_to_phase_defs(self):
        pdefs = cake.to_phase_defs(['p,P', cake.PhaseDef('PP')])
        assert len(pdefs) == 3
        for pdef in pdefs:
            assert isinstance(pdef, cake.PhaseDef)

        pdefs = cake.to_phase_defs(cake.PhaseDef('PP'))
        assert len(pdefs) == 1
        for pdef in pdefs:
            assert isinstance(pdef, cake.PhaseDef)

        pdefs = cake.to_phase_defs('P,p')
        assert len(pdefs) == 2
        for pdef in pdefs:
            assert isinstance(pdef, cake.PhaseDef)

    def test_model_io(self):
        mod = cake.load_model()
        s = cake.write_nd_model_str(mod)
        assert isinstance(s, str)

    def test_dump(self):
        f = BytesIO()
        mod = cake.load_model()
        mod.profile('vp').dump(f)
        f.close()

    def test_angles(self):
        mod = cake.load_model()
        data = [
            [1.0*km, 1.0*km, 1.0*km, 90., 90., 'P'],
            [1.0*km, 2.0*km, 1.0*km, 45., 135., 'P\\'],
            [2.0*km, 1.0*km, 1.0*km, 135., 45., 'p'],
            [1.0*km, 2.0*km, math.sqrt(3.)*km, 60., 120., 'P\\'],
            [2.0*km, 1.0*km, math.sqrt(3.)*km, 120., 60., 'p']]

        for (zstart, zstop, dist, takeoff_want, incidence_want, pdef_want) \
                in data:

            rays = mod.arrivals(
                zstart=zstart,
                zstop=zstop,
                phases=[cake.PhaseDef(sphase)
                        for sphase in 'P,p,P\\,p\\'.split(',')],
                distances=[dist*cake.m2d])

            for ray in rays:
                takeoff = round(ray.takeoff_angle())
                incidence = round(ray.incidence_angle())
                pdef = ray.used_phase().definition()

                assert takeoff == takeoff_want
                assert incidence == incidence_want
                assert pdef == pdef_want


if __name__ == "__main__":
    util.setup_logging('test_cake', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import range

import random
import math
import unittest
import logging
from tempfile import mkdtemp
import shutil

import numpy as num

from .common import Benchmark
from multiprocessing import cpu_count

from pyrocko import util, trace, gf, cake  # noqa
from pyrocko.fomosto import qseis, ahfullgreen

logger = logging.getLogger('pyrocko.test.test_gf_qseis')
benchmark = Benchmark()

km = 1000.

r2d = 180. / math.pi
d2r = 1.0 / r2d


def rand(mi, ma):
    return mi + random.random() * (ma-mi)


def g(trs, cha):
    for tr in trs:
        if tr.channel == cha:
            return tr


@unittest.skipUnless(
    qseis.have_backend(), 'backend qseis not available')
class GFQSeisTestCase(unittest.TestCase):

    tempdirs = []

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def test_pyrocko_gf_vs_qseis(self):
        random.seed(2017)

        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
mantle
 35. 8.04 4.48 3.58 1449. 600.
 77.5 8.045 4.49 3.5 1445. 600.
 77.5 8.045 4.49 3.5 180.6 75.
 120. 8.05 4.5 3.427 180. 75.
 120. 8.05 4.5 3.427 182.6 76.06
 165. 8.175 4.509 3.371 188.7 76.55
 210. 8.301 4.518 3.324 201. 79.4
 210. 8.3 4.52 3.321 336.9 133.3
 410. 9.03 4.871 3.504 376.5 146.1
 410. 9.36 5.08 3.929 414.1 162.7
 660. 10.2 5.611 3.918 428.5 172.9
 660. 10.79 5.965 4.229 1349. 549.6'''.lstrip()))

        store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(store_dir)

        qsconf = qseis.QSeisConfig()
        qsconf.qseis_version = '2006a'

        qsconf.time_region = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qsconf.cut = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qsconf.wavelet_duration_samples = 0.001
        qsconf.sw_flat_earth_transform = 0

        config = gf.meta.ConfigTypeA(
            id='qseis_test',
            sample_rate=0.25,
            receiver_depth=0.*km,
            source_depth_min=10*km,
            source_depth_max=10*km,
            source_depth_delta=1*km,
            distance_min=550*km,
            distance_max=560*km,
            distance_delta=1*km,
            modelling_code_id='qseis.2006a',
            earthmodel_1d=mod,
            tabulated_phases=[
                gf.meta.TPDef(
                    id='begin',
                    definition='p,P,p\\,P\\'),
                gf.meta.TPDef(
                    id='end',
                    definition='2.5'),
            ])

        config.validate()
        gf.store.Store.create_editables(
            store_dir, config=config, extra={'qseis': qsconf})

        store = gf.store.Store(store_dir, 'r')
        store.make_ttt()
        store.close()

        try:
            qseis.build(store_dir, nworkers=1)
        except qseis.QSeisError as e:
            if str(e).find('could not start qseis') != -1:
                logger.warn('qseis not installed; '
                            'skipping test_pyrocko_gf_vs_qseis')
                return
            else:
                raise

        source = gf.MTSource(
            lat=0.,
            lon=0.,
            depth=10.*km)

        source.m6 = tuple(random.random()*2.-1. for x in range(6))

        azi = random.random()*365.
        dist = 553.*km

        dnorth = dist * math.cos(azi*d2r)
        deast = dist * math.sin(azi*d2r)

        targets = []
        for cha in 'rtz':
            target = gf.Target(
                quantity='displacement',
                codes=('', '0000', 'PG', cha),
                north_shift=dnorth,
                east_shift=deast,
                depth=config.receiver_depth,
                store_id='qseis_test')

            dist = source.distance_to(target)
            azi, bazi = source.azibazi_to(target)

            if cha == 'r':
                target.azimuth = bazi + 180.
                target.dip = 0.
            elif cha == 't':
                target.azimuth = bazi - 90.
                target.dip = 0.
            elif cha == 'z':
                target.azimuth = 0.
                target.dip = 90.

            targets.append(target)

        runner = qseis.QSeisRunner()
        conf = qseis.QSeisConfigFull()
        conf.qseis_version = '2006a'
        conf.receiver_distances = [dist/km]
        conf.receiver_azimuths = [azi]
        conf.source_depth = source.depth/km
        conf.time_start = 0.0
        conf.time_window = 508.
        conf.time_reduction_velocity = 0.0
        conf.nsamples = 128
        conf.source_mech = qseis.QSeisSourceMechMT(
            mnn=source.mnn,
            mee=source.mee,
            mdd=source.mdd,
            mne=source.mne,
            mnd=source.mnd,
            med=source.med)
        conf.earthmodel_1d = mod

        conf.sw_flat_earth_transform = 0

        runner.run(conf)

        trs = runner.get_traces()
        for tr in trs:
            tr.shift(-config.deltat)
            tr.snap(interpolate=True)
            tr.lowpass(4, 0.05)
            tr.highpass(4, 0.01)

        engine = gf.LocalEngine(store_dirs=[store_dir])

        def process_wrap(nthreads=0):
            @benchmark.labeled('pyrocko.gf.process (nthreads-%d)' % nthreads)
            def process(nthreads):
                return engine.process(source, targets, nthreads=nthreads)\
                    .pyrocko_traces()
            return process(nthreads)

        for nthreads in range(1, cpu_count()+1):
            trs2 = process_wrap(nthreads)
        # print benchmark

        for tr in trs2:
            tr.snap(interpolate=True)
            tr.lowpass(4, 0.05)
            tr.highpass(4, 0.01)

        # trace.snuffle(trs+trs2)

        for cha in 'rtz':
            t1 = g(trs, cha)
            t2 = g(trs2, cha)
            tmin = max(t1.tmin, t2.tmin)
            tmax = min(t1.tmax, t2.tmax)
            t1.chop(tmin, tmax)
            t2.chop(tmin, tmax)
            d = 2.0 * num.sum((t1.ydata - t2.ydata)**2) / \
                (num.sum(t1.ydata**2) + num.sum(t2.ydata**2))

            assert d < 0.05

    def test_qseis_vs_ahfull(self):
        random.seed(23)

        vp = 5.8 * km
        vs = 3.46 * km

        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
  0. %(vp)g %(vs)g 2.6 1264. 600.
 20. %(vp)g %(vs)g 2.6 1264. 600.'''.lstrip() % dict(vp=vp/km, vs=vs/km)))

        store_id_qseis = 'homogeneous_qseis'
        store_id_ahfull = 'homogeneous_ahfull'

        qsconf = qseis.QSeisConfig()
        qsconf.qseis_version = '2006a'

        textra = 5.0

        qsconf.time_region = (
            gf.meta.Timing('{vel:%g}-%g' % (vp/km, textra)),
            gf.meta.Timing('{vel:%g}+%g' % (vs/km, textra)))

        qsconf.cut = (
            gf.meta.Timing('{vel:%g}-%g' % (vp/km, textra)),
            gf.meta.Timing('{vel:%g}+%g' % (vs/km, textra)))

        qsconf.relevel_with_fade_in = True

        qsconf.fade = (
            gf.meta.Timing('{vel:%g}-%g' % (vp/km, textra)),
            gf.meta.Timing('{vel:%g}-%g' % (vp/km, 0.)),
            gf.meta.Timing('{vel:%g}+%g' % (vs/km, 0.)),
            gf.meta.Timing('{vel:%g}+%g' % (vs/km, textra)))

        qsconf.wavelet_duration_samples = 0.001
        qsconf.sw_flat_earth_transform = 0
        qsconf.filter_surface_effects = 1
        qsconf.wavenumber_sampling = 5.
        qsconf.aliasing_suppression_factor = 0.01

        sample_rate = 10.

        config = gf.meta.ConfigTypeA(
            id=store_id_qseis,
            sample_rate=sample_rate,
            receiver_depth=0.*km,
            source_depth_min=1.*km,
            source_depth_max=19*km,
            source_depth_delta=6.*km,
            distance_min=2.*km,
            distance_max=20*km,
            distance_delta=2*km,
            modelling_code_id='qseis.2006a',
            earthmodel_1d=mod,
            tabulated_phases=[
                gf.meta.TPDef(
                    id='begin',
                    definition='p,P,p\\,P\\'),
                gf.meta.TPDef(
                    id='end',
                    definition='s,S,s\\,S\\'),
            ])

        config.validate()

        store_dir_qseis = mkdtemp(prefix=store_id_qseis)
        self.tempdirs.append(store_dir_qseis)

        gf.store.Store.create_editables(
            store_dir_qseis, config=config, extra={'qseis': qsconf})

        store = gf.store.Store(store_dir_qseis, 'r')
        store.make_ttt()
        store.close()

        try:
            qseis.build(store_dir_qseis, nworkers=1)
        except qseis.QSeisError as e:
            if str(e).find('could not start qseis') != -1:
                logger.warn('qseis not installed; '
                            'skipping test_pyrocko_gf_vs_qseis')
                return
            else:
                raise

        config = gf.meta.ConfigTypeA(
            id=store_id_ahfull,
            sample_rate=sample_rate,
            receiver_depth=0.*km,
            source_depth_min=1.*km,
            source_depth_max=19*km,
            source_depth_delta=6.*km,
            distance_min=2.*km,
            distance_max=20*km,
            distance_delta=2*km,
            modelling_code_id='ahfullgreen',
            earthmodel_1d=mod,
            tabulated_phases=[
                gf.meta.TPDef(
                    id='begin',
                    definition='p,P,p\\,P\\'),
                gf.meta.TPDef(
                    id='end',
                    definition='s,S,s\\,S\\'),
            ])

        config.validate()

        store_dir_ahfull = mkdtemp(prefix=store_id_qseis)
        self.tempdirs.append(store_dir_ahfull)

        gf.store.Store.create_editables(
            store_dir_ahfull, config=config)

        store = gf.store.Store(store_dir_ahfull, 'r')
        store.make_ttt()
        store.close()

        ahfullgreen.build(store_dir_ahfull, nworkers=1)

        sdepth = rand(config.source_depth_min, config.source_depth_max)
        sdepth = round(
            (sdepth - config.source_depth_min)
            / config.source_depth_delta) * config.source_depth_delta \
            + config.source_depth_min

        source = gf.MTSource(
            lat=0.,
            lon=0.,
            depth=sdepth)

        source.m6 = tuple(rand(-1., 1.) for x in range(6))

        for ii in range(5):
            azi = random.random()*365.
            dist = rand(config.distance_min, config.distance_max)
            dist = round(dist / config.distance_delta) * config.distance_delta

            dnorth = dist * math.cos(azi*d2r)
            deast = dist * math.sin(azi*d2r)

            targets = []
            for cha in 'rtz':
                target = gf.Target(
                    quantity='displacement',
                    codes=('', '0000', 'PG', cha),
                    north_shift=dnorth,
                    east_shift=deast,
                    depth=config.receiver_depth,
                    store_id=store_id_ahfull)

                dist = source.distance_to(target)
                azi, bazi = source.azibazi_to(target)

                if cha == 'r':
                    target.azimuth = bazi + 180.
                    target.dip = 0.
                elif cha == 't':
                    target.azimuth = bazi - 90.
                    target.dip = 0.
                elif cha == 'z':
                    target.azimuth = 0.
                    target.dip = 90.

                targets.append(target)

            runner = qseis.QSeisRunner()
            conf = qseis.QSeisConfigFull()
            conf.qseis_version = '2006a'
            conf.receiver_distances = [dist/km]
            conf.receiver_azimuths = [azi]
            conf.receiver_depth = config.receiver_depth / km
            conf.source_depth = source.depth / km

            distance_3d_max = math.sqrt(
                config.distance_max**2 + (
                    config.source_depth_max - config.source_depth_min)**2)

            nsamples = trace.nextpow2(
                int(math.ceil(
                    distance_3d_max / vs * 2.0 + 2.*textra)
                    * config.sample_rate))

            conf.time_start = -textra
            conf.time_window = (nsamples-1) / config.sample_rate
            conf.time_reduction_velocity = 0.0
            conf.nsamples = nsamples
            conf.source_mech = qseis.QSeisSourceMechMT(
                mnn=source.mnn,
                mee=source.mee,
                mdd=source.mdd,
                mne=source.mne,
                mnd=source.mnd,
                med=source.med)
            conf.earthmodel_1d = mod

            conf.sw_flat_earth_transform = 0
            conf.filter_surface_effects = 1
            conf.wavenumber_sampling = 10.
            conf.wavelet_duration_samples = 0.001
            conf.aliasing_suppression_factor = 0.01

            conf.validate()

            runner.run(conf)

            trs = runner.get_traces()
            for tr in trs:
                pass
                tr.lowpass(4, config.sample_rate / 8., demean=False)
                tr.highpass(4, config.sample_rate / 80.)

            engine = gf.LocalEngine(store_dirs=[
                store_dir_ahfull, store_dir_qseis])

            trs2 = engine.process(source, targets).pyrocko_traces()
            for tr in trs2:
                tr.shift(config.deltat)
                tr.lowpass(4, config.sample_rate / 8., demean=False)
                tr.highpass(4, config.sample_rate / 80.)

            # trace.snuffle(trs+trs2)

            tmin = store.t(
                '{vel:%g}' % (vp/km), source, target) - textra*0.2
            tmax = store.t(
                '{vel:%g}' % (vs/km), source, target) + textra*0.2

            for tr in trs + trs2:
                tr.chop(tmin, tmax)

            denom = 0.0
            for cha in 'rtz':
                t1 = g(trs, cha)
                t2 = g(trs2, cha)
                denom += num.sum(t1.ydata**2) + num.sum(t2.ydata**2)

            ds = []
            for cha in 'rtz':
                t1 = g(trs, cha)
                t2 = g(trs2, cha)
                ds.append(2.0 * num.sum((t1.ydata - t2.ydata)**2) / denom)

            ds = num.array(ds)

            # if not num.all(ds < 0.05):
            #    trace.snuffle(trs+trs2)

            assert num.all(ds < 0.05)


if __name__ == '__main__':
    util.setup_logging('test_gf_qseis', 'warning')
    unittest.main()



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
from __future__ import absolute_import, division

from builtins import range
import math
import copy
import logging
import numpy as num

from pyrocko import orthodrome
from pyrocko.orthodrome import wrap
from pyrocko.guts import Object, Float, String, List

from .location import Location

logger = logging.getLogger('pyrocko.model.station')

guts_prefix = 'pf'

d2r = num.pi / 180.


class ChannelsNotOrthogonal(Exception):
    pass


def guess_azimuth_from_name(channel_name):
    if channel_name.endswith('N'):
        return 0.
    elif channel_name.endswith('E'):
        return 90.
    elif channel_name.endswith('Z'):
        return 0.

    return None


def guess_dip_from_name(channel_name):
    if channel_name.endswith('N'):
        return 0.
    elif channel_name.endswith('E'):
        return 0.
    elif channel_name.endswith('Z'):
        return -90.

    return None


def guess_azimuth_dip_from_name(channel_name):
    return guess_azimuth_from_name(channel_name), \
        guess_dip_from_name(channel_name)


def mkvec(x, y, z):
    return num.array([x, y, z], dtype=num.float)


def are_orthogonal(enus, eps=0.05):
    return all(abs(x) < eps for x in [
        num.dot(enus[0], enus[1]),
        num.dot(enus[1], enus[2]),
        num.dot(enus[2], enus[0])])


def fill_orthogonal(enus):

    nmiss = sum(x is None for x in enus)

    if nmiss == 1:
        for ic in range(len(enus)):
            if enus[ic] is None:
                enus[ic] = num.cross(enus[(ic-2) % 3], enus[(ic-1) % 3])

    if nmiss == 2:
        for ic in range(len(enus)):
            if enus[ic] is not None:
                xenu = enus[ic] + mkvec(1, 1, 1)
                enus[(ic+1) % 3] = num.cross(enus[ic], xenu)
                enus[(ic+2) % 3] = num.cross(enus[ic], enus[(ic+1) % 3])

    if nmiss == 3:
        # holy camoly..
        enus[0] = mkvec(1, 0, 0)
        enus[1] = mkvec(0, 1, 0)
        enus[2] = mkvec(0, 0, 1)


class Channel(Object):
    name = String.T()
    azimuth = Float.T(optional=True)
    dip = Float.T(optional=True)
    gain = Float.T(default=1.0)

    def __init__(self, name, azimuth=None, dip=None, gain=1.0):
        if azimuth is None:
            azimuth = guess_azimuth_from_name(name)
        if dip is None:
            dip = guess_dip_from_name(name)

        Object.__init__(
            self,
            name=name,
            azimuth=float_or_none(azimuth),
            dip=float_or_none(dip),
            gain=float(gain))

    @property
    def ned(self):
        if None in (self.azimuth, self.dip):
            return None

        n = math.cos(self.azimuth*d2r)*math.cos(self.dip*d2r)
        e = math.sin(self.azimuth*d2r)*math.cos(self.dip*d2r)
        d = math.sin(self.dip*d2r)
        return mkvec(n, e, d)

    @property
    def enu(self):
        if None in (self.azimuth, self.dip):
            return None

        n = math.cos(self.azimuth*d2r)*math.cos(self.dip*d2r)
        e = math.sin(self.azimuth*d2r)*math.cos(self.dip*d2r)
        d = math.sin(self.dip*d2r)
        return mkvec(e, n, -d)

    def __str__(self):
        return '%s %f %f %g' % (self.name, self.azimuth, self.dip, self.gain)


class Station(Location):
    network = String.T()
    station = String.T()
    location = String.T()
    name = String.T(default='')
    channels = List.T(Channel.T())

    def __init__(self, network='', station='', location='', lat=0.0, lon=0.0,
                 elevation=0.0, depth=0.0, name='', channels=None):

        Location.__init__(
            self,
            network=network, station=station, location=location,
            lat=float(lat), lon=float(lon),
            elevation=elevation and float(elevation) or 0.0,
            depth=depth and float(depth) or 0.0,
            name=name or '',
            channels=channels or [])

        self.dist_deg = None
        self.dist_m = None
        self.azimuth = None
        self.backazimuth = None

    def copy(self):
        return copy.deepcopy(self)

    def set_event_relative_data(self, event, distance_3d=False):
        surface_dist = orthodrome.distance_accurate50m(event, self)
        if distance_3d:
            dd = event.depth - self.depth
            self.dist_m = math.sqrt(dd**2 + surface_dist**2)
        else:
            self.dist_m = surface_dist

        self.dist_deg = surface_dist / orthodrome.earthradius_equator * \
            orthodrome.r2d
        self.azimuth = orthodrome.azimuth(event, self)
        self.backazimuth = orthodrome.azimuth(self, event)

    def set_channels_by_name(self, *args):
        self.set_channels([])
        for name in args:
            self.add_channel(Channel(name))

    def set_channels(self, channels):
        self.channels = []
        for ch in channels:
            self.add_channel(ch)

    def get_channels(self):
        return list(self.channels)

    def get_channel_names(self):
        return set(ch.name for ch in self.channels)

    def remove_channel_by_name(self, name):
        todel = [ch for ch in self.channels if ch.name == name]
        for ch in todel:
            self.channels.remove(ch)

    def add_channel(self, channel):
        self.remove_channel_by_name(channel.name)
        self.channels.append(channel)
        self.channels.sort(key=lambda ch: ch.name)

    def get_channel(self, name):
        for ch in self.channels:
            if ch.name == name:
                return ch

        return None

    def rotation_ne_to_rt(self, in_channel_names, out_channel_names):

        angle = wrap(self.backazimuth + 180., -180., 180.)
        in_channels = [self.get_channel(name) for name in in_channel_names]
        out_channels = [
            Channel(out_channel_names[0],
                    wrap(self.backazimuth+180., -180., 180.),  0., 1.),
            Channel(out_channel_names[1],
                    wrap(self.backazimuth+270., -180., 180.),  0., 1.)]
        return angle, in_channels, out_channels

    def _projection_to(
            self, to, in_channel_names, out_channel_names, use_gains=False):

        in_channels = [self.get_channel(name) for name in in_channel_names]

        # create orthogonal vectors for missing components, such that this
        # won't break projections when components are missing.

        vecs = []
        for ch in in_channels:
            if ch is None:
                vecs.append(None)
            else:
                vec = getattr(ch, to)
                if use_gains:
                    vec /= ch.gain
                vecs.append(vec)

        fill_orthogonal(vecs)
        if not are_orthogonal(vecs):
            raise ChannelsNotOrthogonal(
                'components are not orthogonal: station %s.%s.%s, '
                'channels %s, %s, %s'
                % (self.nsl() + tuple(in_channel_names)))

        m = num.hstack([vec2[:, num.newaxis] for vec2 in vecs])

        m = num.where(num.abs(m) < num.max(num.abs(m))*1e-16, 0., m)

        if to == 'ned':
            out_channels = [
                Channel(out_channel_names[0], 0.,   0., 1.),
                Channel(out_channel_names[1], 90.,  0., 1.),
                Channel(out_channel_names[2], 0.,  90., 1.)]

        elif to == 'enu':
            out_channels = [
                Channel(out_channel_names[0], 90.,  0., 1.),
                Channel(out_channel_names[1], 0.,   0., 1.),
                Channel(out_channel_names[2], 0., -90., 1.)]

        return m, in_channels, out_channels

    def guess_channel_groups(self):
        cg = {}
        for channel in self.get_channels():
            if len(channel.name) >= 1:
                kind = channel.name[:-1]
                if kind not in cg:
                    cg[kind] = []
                cg[kind].append(channel.name[-1])

        def allin(a, b):
            return all(x in b for x in a)

        out_groups = []
        for kind, components in cg.items():
            for sys in ('ENZ', '12Z', 'XYZ', 'RTZ'):
                if allin(sys, components):
                    out_groups.append(tuple([kind+c for c in sys]))

        return out_groups

    def guess_projections_to_enu(self, out_channels=('E', 'N', 'U'), **kwargs):
        proj = []
        for cg in self.guess_channel_groups():
            try:
                proj.append(self.projection_to_enu(
                    cg, out_channels=out_channels, **kwargs))

            except ChannelsNotOrthogonal as e:
                logger.warning(str(e))

        return proj

    def guess_projections_to_rtu(
            self, out_channels=('R', 'T', 'U'), backazimuth=None, **kwargs):

        if backazimuth is None:
            backazimuth = self.backazimuth
        out_channels_ = [
            Channel(
                out_channels[0], wrap(backazimuth+180., -180., 180.),  0., 1.),
            Channel(
                out_channels[1], wrap(backazimuth+270., -180., 180.),  0., 1.),
            Channel(
                out_channels[2], 0.,  -90., 1.)]

        proj = []
        for (m, in_channels, _) in self.guess_projections_to_enu(**kwargs):
            phi = (backazimuth + 180.)*d2r
            r = num.array([[math.sin(phi),  math.cos(phi), 0.0],
                           [math.cos(phi), -math.sin(phi), 0.0],
                           [0.0, 0.0, 1.0]])
            proj.append((num.dot(r, m), in_channels, out_channels_))

        return proj

    def projection_to_enu(
            self,
            in_channels,
            out_channels=('E', 'N', 'U'),
            **kwargs):

        return self._projection_to('enu', in_channels, out_channels, **kwargs)

    def projection_to_ned(
            self,
            in_channels,
            out_channels=('N', 'E', 'D'),
            **kwargs):

        return self._projection_to('ned', in_channels, out_channels, **kwargs)

    def projection_from_enu(
            self,
            in_channels=('E', 'N', 'U'),
            out_channels=('X', 'Y', 'Z'),
            **kwargs):

        m, out_channels, in_channels = self._projection_to(
            'enu', out_channels, in_channels, **kwargs)

        return num.linalg.inv(m), in_channels, out_channels

    def projection_from_ned(
            self,
            in_channels=('N', 'E', 'D'),
            out_channels=('X', 'Y', 'Z'),
            **kwargs):

        m, out_channels, in_channels = self._projection_to(
            'ned', out_channels, in_channels, **kwargs)

        return num.linalg.inv(m), in_channels, out_channels

    def nsl_string(self):
        return '.'.join((self.network, self.station, self.location))

    def nsl(self):
        return self.network, self.station, self.location

    def oldstr(self):
        nsl = '%s.%s.%s' % (self.network, self.station, self.location)
        s = '%-15s  %14.5f %14.5f %14.1f %14.1f %s' % (
            nsl, self.lat, self.lon, self.elevation, self.depth, self.name)
        return s


def dump_stations(stations, filename):
    '''Write stations file.

    :param stations: list of :py:class:`Station` objects
    :param filename: filename as str
    '''
    f = open(filename, 'w')
    for sta in stations:
        f.write(sta.oldstr()+'\n')
        for cha in sta.get_channels():
            azimuth = 'NaN'
            if cha.azimuth is not None:
                azimuth = '%7g' % cha.azimuth

            dip = 'NaN'
            if cha.dip is not None:
                dip = '%7g' % cha.dip

            f.write('%5s %14s %14s %14g\n' % (
                cha.name, azimuth, dip, cha.gain))

    f.close()


def float_or_none(s):
    if s is None:
        return None
    elif isinstance(s, str) and s.lower() == 'nan':
        return None
    else:
        return float(s)


def load_stations(filename):
    '''Read stations file.

    :param filename: filename
    :returns: list of :py:class:`Station` objects
    '''
    stations = []
    f = open(filename, 'r')
    station = None
    channel_names = []
    for (iline, line) in enumerate(f):
        toks = line.split(None, 5)
        if line.strip().startswith('#') or line.strip() == '':
            continue

        if len(toks) == 5 or len(toks) == 6:
            net, sta, loc = toks[0].split('.')
            lat, lon, elevation, depth = [float(x) for x in toks[1:5]]
            if len(toks) == 5:
                name = ''
            else:
                name = toks[5].rstrip()

            station = Station(
                net, sta, loc, lat, lon,
                elevation=elevation, depth=depth, name=name)

            stations.append(station)
            channel_names = []

        elif len(toks) == 4 and station is not None:
            name, azi, dip, gain = (
                toks[0],
                float_or_none(toks[1]),
                float_or_none(toks[2]),
                float(toks[3]))
            if name in channel_names:
                logger.warning(
                    'redefined channel! (line: %i, file: %s)' %
                    (iline + 1, filename))
            else:
                channel_names.append(name)

            channel = Channel(name, azimuth=azi, dip=dip, gain=gain)
            station.add_channel(channel)

        else:
            logger.warning('skipping invalid station/channel definition '
                           '(line: %i, file: %s' % (iline + 1, filename))

    f.close()
    return stations


def dump_kml(objects, filename):
    station_template = '''
  <Placemark>
    <name>%(network)s.%(station)s.%(location)s</name>
    <description></description>
    <styleUrl>#msn_S</styleUrl>
    <Point>
      <coordinates>%(lon)f,%(lat)f,%(elevation)f</coordinates>
    </Point>
  </Placemark>
'''

    f = open(filename, 'w')
    f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    f.write('<kml xmlns="http://www.opengis.net/kml/2.2">\n')
    f.write('<Document>\n')
    f.write(''' <Style id="sh_S">
                <IconStyle>
                        <scale>1.3</scale>
                        <Icon>
                                <href>http://maps.google.com/mapfiles/kml/paddle/S.png</href>
                        </Icon>
                        <hotSpot x="32" y="1" xunits="pixels" yunits="pixels"/>
                </IconStyle>
                <ListStyle>
                        <ItemIcon>
                                <href>http://maps.google.com/mapfiles/kml/paddle/S-lv.png</href>
                        </ItemIcon>
                </ListStyle>
        </Style>
        <Style id="sn_S">
                <IconStyle>
                        <scale>1.1</scale>
                        <Icon>
                                <href>http://maps.google.com/mapfiles/kml/paddle/S.png</href>
                        </Icon>
                        <hotSpot x="32" y="1" xunits="pixels" yunits="pixels"/>
                </IconStyle>
                <ListStyle>
                        <ItemIcon>
                                <href>http://maps.google.com/mapfiles/kml/paddle/S-lv.png</href>
                        </ItemIcon>
                </ListStyle>
        </Style>
        <StyleMap id="msn_S">
                <Pair>
                        <key>normal</key>
                        <styleUrl>#sn_S</styleUrl>
                </Pair>
                <Pair>
                        <key>highlight</key>
                        <styleUrl>#sh_S</styleUrl>
                </Pair>
        </StyleMap>
''')
    for obj in objects:

        if isinstance(obj, Station):
            f.write(station_template % obj.__dict__)
    f.write('</Document>')
    f.write('</kml>\n')
    f.close()



from __future__ import division, print_function, absolute_import
from builtins import range

import unittest
import logging
import random
import time
import math
import multiprocessing
import os.path as op
from pyrocko import gf, util, model
from pyrocko import moment_tensor as pmt
from pyrocko import orthodrome as od

logger = logging.getLogger('pyrocko.test.test_gf_scenarios')
km = 1000.
d2r = math.pi/180.

transchan = {
    'e': 'E',
    'n': 'N',
    'u': 'Z'}


def rand(mi, ma):
    mi = float(mi)
    ma = float(ma)
    return random.random()*(ma-mi) + mi


def to_kiwi_source(source):
    from tunguska import source as kiwi_source

    return kiwi_source.Source(
        time=source.time + source.length / source.velocity / 2.,
        depth=source.depth,
        moment=source.get_moment(),
        strike=source.strike,
        dip=source.dip,
        slip_rake=source.rake,
        length_a=(1. - source.nucleation_x)/2. * source.length,
        length_b=(source.nucleation_x - -1.)/2. * source.length,
        width=source.width,
        rise_time=source.stf.duration)


def have_store(store_id):
    engine = gf.get_engine()
    try:
        engine.get_store(store_id)
        return True
    except gf.NoSuchStore:
        return False


def have_kiwi():
    try:
        import tunguska  # noqa
        return True
    except ImportError:
        return False


class GFScenariosTestCase(unittest.TestCase):
    store_id = 'crust2_mf'
    store_id2 = 'chile_70km_crust'

    @unittest.skipUnless(
            have_store(store_id),
            'GF Store "%s" is not available' % store_id)
    def test_regional(self):
        engine = gf.get_engine()

        nsources = 10
        nstations = 10

        print('cache source channels par wallclock seismograms_per_second')
        nprocs_max = multiprocessing.cpu_count()

        for sourcetype, channels in [
                ['point', 'Z'],
                ['point', 'NEZ'],
                ['rect', 'Z'],
                ['rect', 'NEZ']]:

            for nprocs in [1, 2, 4, 8, 16, 32]:
                if nprocs > nprocs_max:
                    continue

                sources = []
                for isource in range(nsources):
                    m = pmt.MomentTensor.random_dc()
                    strike, dip, rake = map(float, m.both_strike_dip_rake()[0])

                    if sourcetype == 'point':
                        source = gf.DCSource(
                            north_shift=rand(-20.*km, 20*km),
                            east_shift=rand(-20.*km, 20*km),
                            depth=rand(10*km, 20*km),
                            strike=strike,
                            dip=dip,
                            rake=rake,
                            magnitude=rand(4.0, 5.0))

                    elif sourcetype == 'rect':
                        source = gf.RectangularSource(
                            north_shift=rand(-20.*km, 20*km),
                            east_shift=rand(-20.*km, 20*km),
                            depth=rand(10*km, 20*km),
                            length=10*km,
                            width=5*km,
                            nucleation_x=0.,
                            nucleation_y=-1.,
                            strike=strike,
                            dip=dip,
                            rake=rake,
                            magnitude=rand(4.0, 5.0))
                    else:
                        assert False

                    sources.append(source)

                targets = []
                for istation in range(nstations):
                    dist = rand(40.*km, 900*km)
                    azi = rand(-180., 180.)

                    north_shift = dist * math.cos(azi*d2r)
                    east_shift = dist * math.sin(azi*d2r)

                    for cha in channels:
                        target = gf.Target(
                            codes=('', 'S%04i' % istation, '', cha),
                            north_shift=north_shift,
                            east_shift=east_shift,
                            quantity='displacement',
                            interpolation='multilinear',
                            # optimization='disable',
                            store_id=GFScenariosTestCase.store_id)

                        targets.append(target)

                ntraces = len(targets) * len(sources)
                for temperature in ['cold', 'hot']:
                    t0 = time.time()
                    resp = engine.process(sources, targets, nprocs=nprocs)
                    # print resp.stats

                    t1 = time.time()
                    duration = t1 - t0
                    sps = ntraces / duration
                    if temperature == 'hot':
                        if nprocs == 1:
                            sps_ref = sps
                        print('%-5s %-6s %-8s %3i %9.3f %12.1f %12.1f' % (
                            temperature, sourcetype, channels, nprocs, t1-t0,
                            sps, sps/sps_ref))

                    del resp

    @unittest.skipUnless(
            have_store(store_id2),
            'GF Store "%s" is not available' % store_id2)
    @unittest.skipUnless(
            have_kiwi(),
            'KIWI Tools not available')
    def test_against_kiwi(self):
        engine = gf.get_engine()
        store_id = GFScenariosTestCase.store_id2
        try:
            store = engine.get_store(store_id)
        except gf.NoSuchStore:
            logger.warn('GF Store %s not available - skipping test' % store_id)
            return

        base_source = gf.RectangularSource(
            depth=15*km,
            strike=0.,
            dip=90.,
            rake=0.,
            magnitude=4.5,
            nucleation_x=-1.,
            length=10*km,
            width=0*km,
            stf=gf.BoxcarSTF(duration=1.0))

        base_event = base_source.pyrocko_event()

        channels = 'NEZ'
        nstations = 20
        stations = []
        targets = []
        for istation in range(nstations):
            dist = rand(40.*km, 900*km)
            azi = rand(-180., 180.)
            north_shift = dist * math.cos(azi*d2r)
            east_shift = dist * math.sin(azi*d2r)
            lat, lon = od.ne_to_latlon(0., 0., north_shift, east_shift)
            sta = 'S%02i' % istation
            station = model.Station(
                '', sta, '',
                lat=lat,
                lon=lon)

            station.set_channels_by_name('N', 'E', 'Z')
            stations.append(station)

            for cha in channels:
                target = gf.Target(
                    codes=station.nsl() + (cha,),
                    lat=lat,
                    lon=lon,
                    quantity='displacement',
                    interpolation='multilinear',
                    # optimization='disable',
                    store_id=store_id)

                targets.append(target)

        from tunguska import glue

        nsources = 10

        # nprocs_max = multiprocessing.cpu_count()
        nprocs = 1

        try:
            seis = glue.start_seismosizer(
                gfdb_path=op.join(store.store_dir, 'db'),
                event=base_event,
                stations=stations,
                hosts=['localhost']*nprocs,
                balance_method='123321',
                effective_dt=0.5,
                verbose=False)

            ksource = to_kiwi_source(base_source)

            seis.set_source(ksource)

            recs = seis.get_receivers_snapshot(('syn',), (), 'plain')
            trs = []
            for rec in recs:
                for tr in rec.get_traces():
                    tr.set_codes(channel=transchan[tr.channel])
                    trs.append(tr)

            # trs2 = engine.process(base_source, targets).pyrocko_traces()
            engine.process(base_source, targets).pyrocko_traces()

            # trace.snuffle(trs + trs2)

            seis.set_synthetic_reference()

            for sourcetype in ['point', 'rect']:
                sources = []
                for isource in range(nsources):
                    m = pmt.MomentTensor.random_dc()
                    strike, dip, rake = map(float, m.both_strike_dip_rake()[0])

                    if sourcetype == 'point':
                        source = gf.RectangularSource(
                            north_shift=rand(-20.*km, 20*km),
                            east_shift=rand(-20.*km, 20*km),
                            depth=rand(10*km, 20*km),
                            nucleation_x=0.0,
                            nucleation_y=0.0,
                            strike=strike,
                            dip=dip,
                            rake=rake,
                            magnitude=rand(4.0, 5.0),
                            stf=gf.BoxcarSTF(duration=1.0))

                    elif sourcetype == 'rect':
                        source = gf.RectangularSource(
                            north_shift=rand(-20.*km, 20*km),
                            east_shift=rand(-20.*km, 20*km),
                            depth=rand(10*km, 20*km),
                            length=10*km,
                            width=5*km,
                            nucleation_x=-1.,
                            nucleation_y=0,
                            strike=strike,
                            dip=dip,
                            rake=rake,
                            magnitude=rand(4.0, 5.0),
                            stf=gf.BoxcarSTF(duration=1.0))
                    else:
                        assert False

                    sources.append(source)

                for temperature in ['cold', 'hot']:
                    t0 = time.time()
                    resp = engine.process(sources, targets, nprocs=nprocs)
                    t1 = time.time()
                    if temperature == 'hot':
                        dur_pyrocko = t1 - t0

                    del resp

                ksources = map(to_kiwi_source, sources)

                for temperature in ['cold', 'hot']:
                    t0 = time.time()
                    seis.make_misfits_for_sources(
                        ksources, show_progress=False)
                    t1 = time.time()
                    if temperature == 'hot':
                        dur_kiwi = t1 - t0

                print('pyrocko %-5s %5.2fs  %5.1fx' % (
                    sourcetype, dur_pyrocko, 1.0))
                print('kiwi    %-5s %5.2fs  %5.1fx' % (
                    sourcetype, dur_kiwi, dur_pyrocko/dur_kiwi))

        finally:
            seis.close()
            del seis


if __name__ == '__main__':
    util.setup_logging('test_gf_scenarios', 'warning')
    unittest.main()
